# -*- coding: utf-8 -*-
"""
http://matplotlib.org/examples/user_interfaces/embedding_in_qt4.html
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import time
from guitool_ibeis.__PYQT__ import QtCore, QtWidgets
from guitool_ibeis import __PYQT__
if __PYQT__.GUITOOL_PYQT_VERSION == 5:
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
elif __PYQT__.GUITOOL_PYQT_VERSION == 4:
    from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
else:
    raise NotImplementedError
from matplotlib.figure import Figure
#from matplotlib.backends import qt_compat


BASE = FigureCanvas


class QtAbstractMplInteraction(BASE):
    """
    Ultimately, this is a QWidget (as well as a FigureCanvasAgg, etc.).

    Args:
        self (?):
        parent (None): (default = None)
        width (int): (default = 5)
        height (int): (default = 4)
        dpi (int): (default = 100)

    CommandLine:
        python -m guitool_ibeis.mpl_embed QtAbstractMplInteraction --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(env:DISPLAY)
        >>> from guitool_ibeis.mpl_embed import *  # NOQA
        >>> import plottool_ibeis as pt
        >>> import guitool_ibeis
        >>> guitool_ibeis.ensure_qapp()  # must be ensured before any embeding
        >>> self = QtAbstractMplInteraction()
        >>> parent = None
        >>> width = 5
        >>> height = 4
        >>> dpi = 100
        >>> self = QtAbstractMplInteraction(parent)
        >>> self.show()
        >>> print('Blocking')
        >>> self.start_blocking()
        >>> print('Done')
        >>> # xdoctest: +REQUIRES(--show)
        >>> #import utool as ut
        >>> #ut.quit_if_noshow()
        >>> #import plottool_ibeis as pt
        >>> #ut.show_if_requested()
        >>> guitool_ibeis.qtapp_loop(self, frequency=100, init_signals=True)

    """

    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi)
        self._running = None
        self.axes = fig.add_subplot(111)
        # We want the axes cleared every time plot() is called
        has_hold = False
        try:
            import matplotlib as mpl
            from distutils.version import LooseVersion
            has_hold = LooseVersion(mpl.__version__) < '3.0.0'
        except Exception:
            pass

        if has_hold:
            self.axes.hold(False)

        self.compute_initial_figure()
        #
        BASE.__init__(self, fig)
        self.setParent(parent)

        BASE.setSizePolicy(self, QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding)
        BASE.updateGeometry(self)
        self.fig = fig

    def compute_initial_figure(self):
        pass

    def closeEvent(self, event):
        self.stop_blocking()
        event.accept()
        #BASE.closeEvent(self, event)

    @QtCore.pyqtSlot()
    def start_blocking(self):
        #self.buttonStart.setDisabled(True)
        self._running = True
        while self._running:
            QtWidgets.qApp.processEvents()
            time.sleep(0.05)
        #self.buttonStart.setDisabled(False)

    @QtCore.pyqtSlot()
    def stop_blocking(self):
        self._running = False


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m guitool_ibeis.mpl_embed
        python -m guitool_ibeis.mpl_embed --allexamples
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# DEPRICATE?
from __future__ import absolute_import, division, print_function
from guitool_ibeis.__PYQT__ import QtCore, QtGui
from guitool_ibeis.__PYQT__ import QtWidgets
from guitool_ibeis.__PYQT__.QtCore import Qt
from guitool_ibeis.guitool_delegates import ComboDelegate, ButtonDelegate
from guitool_ibeis import qtype
import utool
(print, rrr, profile) = utool.inject2(__name__)


class ColumnListTableView(QtWidgets.QTableView):
    """ Table View for an AbstractItemModel """
    def __init__(view, *args, **kwargs):
        super(ColumnListTableView, view).__init__(*args, **kwargs)
        view.setSortingEnabled(True)
        view.vertical_header = view.verticalHeader()
        view.vertical_header.setVisible(True)
        #view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        view.resizeColumnsToContents()

    @QtCore.pyqtSlot()
    def cellButtonClicked(self):
        print(self.sender())
        print(self.sender().text())


class ColumnListItemModel(QtCore.QAbstractTableModel):
    """ Item model for displaying a list of columns """

    #
    # Non-Qt Init Functions

    def __init__(model, col_data_list=None, col_name_list=None, niceheader_list=None,
                 col_type_list=None, col_edit_list=None,
                 display_indices=False, col_sort_index=None,
                 parent=None, *args):
        super(ColumnListItemModel, model).__init__()
        model.sortcolumn = None
        model.sortreverse = False
        model.display_indices = False  # FIXME: Broken
        model._change_data(col_data_list, col_name_list, niceheader_list,
                           col_type_list, col_edit_list, display_indices,
                           col_sort_index)

    def _change_data(model, col_data_list=None, col_name_list=None,
                     niceheader_list=None, col_type_list=None,
                     col_edit_list=None, display_indices=False,
                     col_sort_index=None):
        model.layoutAboutToBeChanged.emit()
        if col_name_list is None:
            col_name_list = []
        if col_data_list is None:
            col_data_list = []
        if len(col_data_list) > 0:
            model.sortcolumn = 0
        #print('[model] Changing data')
        # Set the data to display (list of lists)
        model._change_columns(col_data_list, col_type_list)
        # Set the headers
        model._change_headers(col_name_list, niceheader_list)
        # Is editable
        model._change_editable(col_edit_list)
        # Make internal indices
        model.set_sorting(col_sort_index)
        model.display_indices = display_indices
        model._change_row_indices()
        # Make sure the user didn't do anything bad
        model._assert_feasibility()
        model.layoutChanged.emit()

    def _change_editable(model, col_edit_list=None):
        if col_edit_list is None:
            model.column_editable = [False] * len(model.col_data_list)
        else:
            model.column_editable = [header in col_edit_list for header in
                                     model.col_name_list]
        #print('[model] new column_editable = %r' % (model.column_editable,))

    def _change_headers(model, col_name_list, niceheader_list=None):
        """ Internal header names """
        model.col_name_list = col_name_list
        # Set user readable "nice" headers, default to internal ones
        if niceheader_list is None:
            model.niceheader_list = col_name_list
        else:
            model.niceheader_list = niceheader_list
        #print('[model] new col_name_list = %r' % (model.col_name_list,))

    def _change_columns(model, col_data_list, col_type_list=None):
        model.col_data_list = col_data_list
        if col_type_list is not None:
            model.col_type_list = col_type_list
        else:
            model.col_type_list = qtype.infer_coltype(model.col_data_list)
        #print('[model] new col_type_list = %r' % (model.col_type_list,))

    def _change_row_indices(model):
        """  Non-Qt Helper """
        if model.sortcolumn is not None:
            print('using: sortcolumn=%r' % model.sortcolumn)
            column_data = model.col_data_list[model.sortcolumn]
            indices = list(range(len(column_data)))
            model.row_sortx = utool.sortedby(indices, column_data,
                                             reverse=model.sortreverse)
        elif len(model.col_data_list) > 0:
            model.row_sortx = list(range(len(model.col_data_list[0])))
        else:
            model.row_sortx = []
        #print('[model] new len(row_sortx) = %r' % (len(model.row_sortx),))

    def _assert_feasibility(model):
        assert len(model.col_name_list) == len(model.col_data_list)
        nrows = list(map(len, model.col_data_list))
        assert all([nrows[0] == num for num in nrows]), 'inconsistent data'
        #print('[model] is feasible')

    #
    #
    # Non-Qt Helper function

    def get_data(model, index):
        """ Non-Qt Helper """
        row = index.row()
        column = index.column()
        row_data = model.col_data_list[column]
        data = row_data[model.row_sortx[row]]
        return data

    def set_sorting(model, col_sort_index=None, order=Qt.DescendingOrder):
        model.sortreverse = (order == Qt.DescendingOrder)
        if col_sort_index is not None:
            if utool.is_str(col_sort_index):
                col_sort_index = model.col_name_list.index(col_sort_index)
            assert utool.is_int(col_sort_index), 'sort by an index not %r' % type(col_sort_index)
            model.sortcolumn = col_sort_index
            assert model.sortcolumn < len(model.col_name_list), 'outofbounds'
            print('sortcolumn: %r' % model.sortcolumn)

    def set_data(model, index, data):
        """ Non-Qt Helper """
        row = index.row()
        column = index.column()
        row_data = model.col_data_list[column]
        row_data[model.row_sortx[row]] = data

    def get_header(model, column):
        """ Non-Qt Helper """
        return model.col_name_list[column]

    def get_header_data(model, header, row):
        """ Non-Qt Helper """
        column = model.col_name_list.index(header)
        index  = model.index(row, column)
        data   = model.get_data(index)
        return data

    def get_coltype(model, column):
        type_ = model.col_type_list[column]
        return type_

    def get_editable(model, column):
        return model.column_editable[column]

    def get_niceheader(model, column):
        """ Non-Qt Helper """
        return model.niceheader_list[column]

    def get_column_alignment(model, column):
        coltype = model.get_coltype(column)
        if coltype in utool.VALID_FLOAT_TYPES:
            return Qt.AlignRight
        else:
            return Qt.AlignHCenter

    #
    #
    # Qt AbstractItemTable Overrides

    def rowCount(model, parent=QtCore.QModelIndex()):
        """ Qt Override """
        return len(model.row_sortx)

    def columnCount(model, parent=QtCore.QModelIndex()):
        """ Qt Override """
        return len(model.col_name_list)

    def index(model, row, column, parent=QtCore.QModelIndex()):
        """ Qt Override """
        return model.createIndex(row, column)

    def data(model, index, role=Qt.DisplayRole):
        """ Returns the data to display """
        if not index.isValid():
            return None
        flags = model.flags(index)
        if role == Qt.TextAlignmentRole:
            return model.get_column_alignment(index.column())
        if role == Qt.BackgroundRole and (flags & Qt.ItemIsEditable or
                                          flags & Qt.ItemIsUserCheckable):
            return QtCore.QVariant(QtGui.QColor(250, 240, 240))
        if role == Qt.DisplayRole or role == Qt.CheckStateRole:
            data = model.get_data(index)
            var = qtype.cast_into_qt(data, role, flags)
            return var
        else:
            return QtCore.QVariant()

    def setData(model, index, var, role=Qt.EditRole):
        """ Sets the role data for the item at index to var.
        var is a QVariant (called data in documentation)
        """
        print('[model] setData: %r' % (str(qtype.qindexinfo(index))))
        try:
            if not index.isValid():
                return None
            flags = model.flags(index)
            if not (flags & Qt.ItemIsEditable or flags & Qt.ItemIsUserCheckable):
                return None
            if role == Qt.CheckStateRole:
                type_ = 'QtCheckState'
                data = var == Qt.Checked
            elif role != Qt.EditRole:
                return False
            else:
                # Cast var into datatype
                type_ = model.get_coltype(index.column())
                data = qtype.cast_from_qt(var, type_)
            # Do actual setting of data
            print(' * new_data = %s(%r)' % (utool.type_str(type_), data,))
            model.set_data(index, data)
            # Emit that data was changed and return succcess
            model.dataChanged.emit(index, index)
            return True
        except Exception as ex:
            var_ = str(var.toString())  # NOQA
            utool.printex(ex, 'ignoring setData', '[model]',
                          key_list=['var_'])
            #raise
            #print(' * ignoring setData: %r' % locals().get('var', None))
            return False

    def headerData(model, section, orientation, role=Qt.DisplayRole):
        """ Qt Override """
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return model.get_niceheader(section)
        else:
            return QtCore.QVariant()

    def sort(model, column, order):
        """ Qt Override """
        model.layoutAboutToBeChanged.emit()
        model.set_sorting(column, order)
        model._change_row_indices()
        model.layoutChanged.emit()

    def flags(model, index):
        """ Qt Override """
        #return Qt.ItemFlag(0)
        column = index.column()
        if not model.get_editable(column):
            return Qt.ItemIsEnabled | Qt.ItemIsSelectable
        if model.get_coltype(column) in utool.VALID_BOOL_TYPES:
            return Qt.ItemIsEnabled | Qt.ItemIsUserCheckable
        else:
            return Qt.ItemIsEnabled | Qt.ItemIsEditable | Qt.ItemIsSelectable


class ColumnListTableWidget(QtWidgets.QWidget):
    """ ColumnList Table Main Widget """
    def __init__(cltw, col_data_list=None, col_name_list=None,
                 niceheader_list=None, col_type_list=None,
                 col_edit_list=None, display_indices=False,
                 col_sort_index=None, parent=None):
        super(ColumnListTableWidget, cltw).__init__(parent)
        # QtWidgets.QWidget.__init__(cltw, parent)
        # Create vertical layout for the table to go into
        cltw.vert_layout = QtWidgets.QVBoxLayout(cltw)
        # Instansiate the AbstractItemModel
        cltw.model = ColumnListItemModel(parent=cltw)
        # Create a ColumnListTableView for the AbstractItemModel
        cltw.view = ColumnListTableView(cltw)
        cltw.view.setModel(cltw.model)
        cltw.vert_layout.addWidget(cltw.view)
        # Make sure we don't call a childs method
        ColumnListTableWidget.change_data(cltw, col_data_list, col_name_list,
                                          niceheader_list, col_type_list,
                                          col_edit_list, display_indices,
                                          col_sort_index)

    def change_data(cltw, col_data_list=None, col_name_list=None,
                    niceheader_list=None, col_type_list=None,
                    col_edit_list=None, display_indices=False,
                    col_sort_index=None):
        """
        Checks for deligates
        """
        marked_columns = []  # these will be persistantly editable
        if col_type_list is not None:
            print('cltw.change_data: %r' % len(col_type_list))
            col_type_list = list(col_type_list)
            for column in range(len(col_type_list)):
                if isinstance(col_type_list[column], tuple):
                    delegate_type, coltype = col_type_list[column]
                    col_type_list[column] = coltype
                    if cltw.set_column_as_delegate(column, delegate_type):
                        marked_columns.append(column)
        else:
            print('cltw.change_data: None')
        cltw.model._change_data(col_data_list, col_name_list, niceheader_list,
                                col_type_list, col_edit_list,
                                display_indices, col_sort_index)
        # Set persistant editability after data is changed
        for column in marked_columns:
            cltw.set_column_persistant_editor(column)

    def set_column_as_delegate(cltw, column, delegate_type):
        if delegate_type == 'COMBO':
            print('cltw.set_col_del %r %r' % (column, delegate_type))
            cltw.view.setItemDelegateForColumn(column, ComboDelegate(cltw.view))
            return True
        elif delegate_type == 'BUTTON':
            print('cltw.set_col_del %r %r' % (column, delegate_type))
            cltw.view.setItemDelegateForColumn(column, ButtonDelegate(cltw.view))
            return False

    def set_column_persistant_editor(cltw, column):
        """
        Set each row in a column as persistant
        """
        num_rows = cltw.model.rowCount()
        print('cltw.set_persistant: %r rows' % num_rows)
        for row in range(num_rows):
            index  = cltw.model.index(row, column)
            cltw.view.openPersistentEditor(index)

    def is_index_clickable(cltw, index):
        model = index.model()
        clickable = not (model.flags(index) & QtCore.Qt.ItemIsSelectable)
        return clickable

    def get_index_header_data(cltw, header, index):
        model = index.model()
        return model.get_header_data(header, index.row())


def make_listtable_widget(col_data_list, col_name_list, col_edit_list=None,
                          show=True, raise_=True, on_click=None):
    widget = ColumnListTableWidget(col_data_list, col_name_list,
                                   col_edit_list=col_edit_list)

    def on_doubleclick(index):
        # This is actually a release
        #print('DoubleClicked: ' + str(qtype.qindexinfo(index)))
        pass

    def on_pressed(index):
        #print('Pressed: ' + str(qtype.qindexinfo(index)))
        pass

    def on_activated(index):
        #print('Activated: ' + str(qtype.qindexinfo(index)))
        pass

    if on_click is not None:
        widget.view.clicked.connect(on_click)
    widget.view.doubleClicked.connect(on_doubleclick)
    widget.view.pressed.connect(on_doubleclick)
    widget.view.activated.connect(on_activated)
    widget.setGeometry(20, 50, 600, 800)

    if show:
        widget.show()
    if raise_:
        widget.raise_()
    return widget

#if __name__ == '__main__':
    #import sys
    #app = guitoo.ensure_qtapp()
    #widget = DummyWidget()
    #widget.show()
    #widget.raise_()
    #sys.exit(app.exec_())


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m guitool_ibeis.guitool_tables
        python -m guitool_ibeis.guitool_tables --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



from __future__ import absolute_import, division, print_function
from guitool_ibeis.__PYQT__ import QtCore, QtGui
from guitool_ibeis.__PYQT__ import QtWidgets  # NOQA
from guitool_ibeis.__PYQT__.QtCore import Qt
import math
import utool

import six
utool.noinject(__name__, '[StripProxyModel]', DEBUG=False)

#STRIPE_PROXY_BASE = QtGui.QAbstractProxyModel
#STRIPE_PROXY_BASE = QtGui.QSortFilterProxyModel
try:
    STRIPE_PROXY_BASE = QtGui.QIdentityProxyModel
except Exception:
    STRIPE_PROXY_BASE = QtCore.QIdentityProxyModel

STRIP_PROXY_META_CLASS = utool.makeForwardingMetaclass(
    lambda self: self.sourceModel(),
    ['_set_context_id',
     '_get_context_id',
     '_set_changeblocked',
     '_get_changeblocked',
     '_about_to_change',
     '_change',
     '_update',
     '_rows_updated',
     'name'],
    base_class=STRIPE_PROXY_BASE)

STRIP_PROXY_SIX_BASE = six.with_metaclass(STRIP_PROXY_META_CLASS, STRIPE_PROXY_BASE)


class StripeProxyModel(STRIP_PROXY_SIX_BASE):  # (STRIPE_PROXY_BASE, metaclass=STRIP_PROXY_META_CLASS):
    #__metaclass__ = STRIP_PROXY_META_CLASS

    def __init__(self, parent=None, numduplicates=1):
        STRIPE_PROXY_BASE.__init__(self, parent=parent)
        self._nd = numduplicates

    def rowCount(self, parent=QtCore.QModelIndex()):
        sourceParent = self.mapToSource(parent)
        source_rows = self.sourceModel().rowCount(parent=sourceParent)
        rows = math.ceil(source_rows / self._nd)
        #print('StripeProxyModel.rowCount(): %r %r' % (source_rows, rows))
        return int(rows)

    def columnCount(self, parent=QtCore.QModelIndex()):
        source_cols = self.sourceModel().columnCount(parent=parent)
        cols = self._nd * source_cols
        #print('StripeProxyModel.columnCount(): %r %r' % (source_cols, cols))
        return int(cols)

    def proxy_to_source(self, row, col, parent=QtCore.QModelIndex()):
        source_model = self.sourceModel()
        source_cols = source_model.columnCount(parent=parent)
        r, c, p = row, col, parent
        r2 = int(math.floor(c / source_cols)) + (r * self._nd)
        c2 = c % source_cols
        p2 = p
        return r2, c2, p2

    def source_to_proxy(self, row, col, parent=QtCore.QModelIndex()):
        source_model = self.sourceModel()
        source_cols = source_model.columnCount(parent=parent)
        r, c, p = row, col, parent
        r2 = int(math.floor(r / self._nd))
        c2 = ((r % self._nd) * source_cols) + c
        p2 = p
        return r2, c2, p2

    def mapToSource(self, proxyIndex):
        """ returns index into original model """
        if proxyIndex is None:
            return None
        if proxyIndex.isValid():
            r2, c2, p2 = self.proxy_to_source(proxyIndex.row(), proxyIndex.column())
            #print('StripeProxyModel.mapToSource(): %r %r %r; %r %r %r' % (r, c, p, r2, c2, p2))
            sourceIndex = self.sourceModel().index(r2, c2, parent=p2)  # self.sourceModel().root_node[r2]
        else:
            sourceIndex = QtCore.QModelIndex()
        return sourceIndex

    def mapFromSource(self, sourceIndex):
        """ returns index into proxy model """
        if sourceIndex is None:
            return None
        if sourceIndex.isValid():
            r2, c2, p2 = self.source_to_proxy(sourceIndex.row(), sourceIndex.column(), sourceIndex.parent())
            proxyIndex = self.index(r2, c2, p2)
        else:
            proxyIndex = QtCore.QModelIndex()
        return proxyIndex

    def index(self, row, col, parent=QtCore.QModelIndex()):
        if (row, col) != (-1, -1):
            proxyIndex = self.createIndex(row, col, parent)
        else:
            proxyIndex = QtCore.QModelIndex()
        return proxyIndex

    def data(self, proxyIndex, role=Qt.DisplayRole, **kwargs):
        sourceIndex = self.mapToSource(proxyIndex)
        return self.sourceModel().data(sourceIndex, role, **kwargs)

    def setData(self, proxyIndex, value, role=Qt.EditRole):
        sourceIndex = self.mapToSource(proxyIndex)
        return self.sourceModel().setData(sourceIndex, value, role)

    def sort(self, column, order):
        source_model = self.sourceModel()
        source_cols = source_model.columnCount()
        if source_cols > 0:
            source_model.sort(column % source_cols, order)

    def parent(self, index):
        return self.sourceModel().parent(self.mapToSource(index))

#    def mapSelectionToSource(self, sel):

#    def flags(self, *args, **kwargs):
#        return self.sourceModel().flags(*args, **kwargs)

#    def headerData(self, *args, **kwargs):
#        return self.sourceModel().headerData(*args, **kwargs)
#
#    def hasChildren(self, *args, **kwargs):
#        return self.sourceModel().hasChildren(*args, **kwargs)
#
#    def itemData(self, *args, **kwargs):
#        return self.sourceModel().itemData(*args, **kwargs)

    def _update_rows(self):
        return self.sourceModel()._update_rows()

    def _get_row_id(self, proxyIndex):
        return self.sourceModel()._get_row_id(self.mapToSource(proxyIndex))

    def _get_level(self, proxyIndex):
        return self.sourceModel()._get_level(self.mapToSource(proxyIndex))

    def _get_adjacent_qtindex(self, proxyIndex, *args, **kwargs):
        qtindex = self.mapToSource(proxyIndex)
        next_qtindex = self.sourceModel()._get_adjacent_qtindex(qtindex, *args, **kwargs)
        next_proxyindex = self.mapFromSource(next_qtindex)
        return next_proxyindex



# -*- coding: utf-8 -*-
# flake8:noqa
# Wrapper around PyQt4/5
from __future__ import absolute_import, division, print_function
import utool as ut
ut.noinject(__name__, '[__PYQT__.QtTest]')
from . import _internal

if _internal.GUITOOL_PYQT_VERSION == 4:
    from PyQt4.QtTest import *
else:
    from PyQt5.QtTest import *



# flake8:noqa
# Wrapper around PyQt4/5
"""
Move to PyQt5?
    pip install git+git://github.com/pyqt/python-qt5.git

Ignore:
    >>> import sys
    >>> from PyQt5 import QtWidgets
    >>> app = QtWidgets.QApplication(sys.argv)
    >>> button = QtWidgets.QPushButton("Hello")
    >>> button.setFixedSize(400, 400)
    >>> button.show()
    >>> app.exec_()
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import utool as ut
import six
ut.noinject(__name__, '[__PYQT__._internal]')


# SIP must be imported before any PyQt
if six.PY2:
    try:
        import sip
        # http://stackoverflow.com/questions/21217399/pyqt4-qtcore-qvariant-object-instead-of-a-string
        sip.setapi('QVariant', 2)
        sip.setapi('QString', 2)
        sip.setapi('QTextStream', 2)
        sip.setapi('QTime', 2)
        sip.setapi('QUrl', 2)
        sip.setapi('QDate', 2)
        sip.setapi('QDateTime', 2)
        if hasattr(sip, 'setdestroyonexit'):
            sip.setdestroyonexit(False)  # This prevents a crash on windows
    except ImportError as ex:
        print('Warning: Import Error: %s' % str(ex))
    except ValueError as ex:
        print('Warning: Value Error: %s' % str(ex))

try:
    import PyQt5
    GUITOOL_PYQT_VERSION = 5
except ImportError:
    import PyQt4
    GUITOOL_PYQT_VERSION = 4

"""
import __PYQT__
from __PYQT__ import QtCore
from __PYQT__ import QtGui
from __PYQT__ import QtTest
from __PYQT__.QtCore import Qt
from __PYQT__.QtGui import QSizePolicy

python -c "from guitool_ibeis.__PYQT__ import QtCore"
python -c "from guitool_ibeis import __PYQT__"
python -c "import guitool_ibeis.__PYQT__"
python -c "import __PYQT__"
from __PYQT__ import QtGui
from __PYQT__ import QtTest
from __PYQT__.QtCore import Qt
from __PYQT__.QtGui import QSizePolicy
"""



#!/usr/bin/env python
# Generated by ~/code/xcookie/xcookie/builders/setup.py
# based on part ~/code/xcookie/xcookie/rc/setup.py.in
import sys
import re
from os.path import exists, dirname, join
from setuptools import find_packages
from setuptools import setup


def parse_version(fpath):
    """
    Statically parse the version number from a python file
    """
    value = static_parse("__version__", fpath)
    return value


def static_parse(varname, fpath):
    """
    Statically parse the a constant variable from a python file
    """
    import ast

    if not exists(fpath):
        raise ValueError("fpath={!r} does not exist".format(fpath))
    with open(fpath, "r") as file_:
        sourcecode = file_.read()
    pt = ast.parse(sourcecode)

    class StaticVisitor(ast.NodeVisitor):
        def visit_Assign(self, node):
            for target in node.targets:
                if getattr(target, "id", None) == varname:
                    self.static_value = node.value.s

    visitor = StaticVisitor()
    visitor.visit(pt)
    try:
        value = visitor.static_value
    except AttributeError:
        import warnings

        value = "Unknown {}".format(varname)
        warnings.warn(value)
    return value


def parse_description():
    """
    Parse the description in the README file

    CommandLine:
        pandoc --from=markdown --to=rst --output=README.rst README.md
        python -c "import setup; print(setup.parse_description())"
    """
    readme_fpath = join(dirname(__file__), "README.rst")
    # This breaks on pip install, so check that it exists.
    if exists(readme_fpath):
        with open(readme_fpath, "r") as f:
            text = f.read()
        return text
    return ""


def parse_requirements(fname="requirements.txt", versions=False):
    """
    Parse the package dependencies listed in a requirements file but strips
    specific versioning information.

    Args:
        fname (str): path to requirements file
        versions (bool | str, default=False):
            If true include version specs.
            If strict, then pin to the minimum version.

    Returns:
        List[str]: list of requirements items

    CommandLine:
        python -c "import setup, ubelt; print(ubelt.urepr(setup.parse_requirements()))"
    """
    require_fpath = fname

    def parse_line(line, dpath=""):
        """
        Parse information from a line in a requirements text file

        line = 'git+https://a.com/somedep@sometag#egg=SomeDep'
        line = '-e git+https://a.com/somedep@sometag#egg=SomeDep'
        """
        # Remove inline comments
        comment_pos = line.find(" #")
        if comment_pos > -1:
            line = line[:comment_pos]

        if line.startswith("-r "):
            # Allow specifying requirements in other files
            target = join(dpath, line.split(" ")[1])
            for info in parse_require_file(target):
                yield info
        else:
            # See: https://www.python.org/dev/peps/pep-0508/
            info = {"line": line}
            if line.startswith("-e "):
                info["package"] = line.split("#egg=")[1]
            else:
                if "--find-links" in line:
                    # setuptools doesnt seem to handle find links
                    line = line.split("--find-links")[0]
                if ";" in line:
                    pkgpart, platpart = line.split(";")
                    # Handle platform specific dependencies
                    # setuptools.readthedocs.io/en/latest/setuptools.html
                    # #declaring-platform-specific-dependencies
                    plat_deps = platpart.strip()
                    info["platform_deps"] = plat_deps
                else:
                    pkgpart = line
                    platpart = None

                # Remove versioning from the package
                pat = "(" + "|".join([">=", "==", ">"]) + ")"
                parts = re.split(pat, pkgpart, maxsplit=1)
                parts = [p.strip() for p in parts]

                info["package"] = parts[0]
                if len(parts) > 1:
                    op, rest = parts[1:]
                    version = rest  # NOQA
                    info["version"] = (op, version)
            yield info

    def parse_require_file(fpath):
        dpath = dirname(fpath)
        with open(fpath, "r") as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith("#"):
                    for info in parse_line(line, dpath=dpath):
                        yield info

    def gen_packages_items():
        if exists(require_fpath):
            for info in parse_require_file(require_fpath):
                parts = [info["package"]]
                if versions and "version" in info:
                    if versions == "strict":
                        # In strict mode, we pin to the minimum version
                        if info["version"]:
                            # Only replace the first >= instance
                            verstr = "".join(info["version"]).replace(">=", "==", 1)
                            parts.append(verstr)
                    else:
                        parts.extend(info["version"])
                if not sys.version.startswith("3.4"):
                    # apparently package_deps are broken in 3.4
                    plat_deps = info.get("platform_deps")
                    if plat_deps is not None:
                        parts.append(";" + plat_deps)
                item = "".join(parts)
                yield item

    packages = list(gen_packages_items())
    return packages


# # Maybe use in the future? But has private deps
# def parse_requirements_alt(fpath='requirements.txt', versions='loose'):
#     """
#     Args:
#         versions (str): can be
#             False or "free" - remove all constraints
#             True or "loose" - use the greater or equal (>=) in the req file
#             strict - replace all greater equal with equals
#     """
#     # Note: different versions of pip might have different internals.
#     # This may need to be fixed.
#     from pip._internal.req import parse_requirements
#     from pip._internal.network.session import PipSession
#     requirements = []
#     for req in parse_requirements(fpath, session=PipSession()):
#         if not versions or versions == 'free':
#             req_name = req.requirement.split(' ')[0]
#             requirements.append(req_name)
#         elif versions == 'loose' or versions is True:
#             requirements.append(req.requirement)
#         elif versions == 'strict':
#             part1, *rest = req.requirement.split(';')
#             strict_req = ';'.join([part1.replace('>=', '==')] + rest)
#             requirements.append(strict_req)
#         else:
#             raise KeyError(versions)
#     requirements = [r.replace(' ', '') for r in requirements]
#     return requirements


NAME = "guitool_ibeis"
INIT_PATH = "guitool_ibeis/__init__.py"
VERSION = parse_version(INIT_PATH)

if __name__ == "__main__":
    setupkw = {}

    setupkw["install_requires"] = parse_requirements(
        "requirements/runtime.txt", versions="loose"
    )
    setupkw["extras_require"] = {
        "all": parse_requirements("requirements.txt", versions="loose"),
        "headless": parse_requirements("requirements/headless.txt", versions="loose"),
        "graphics": parse_requirements("requirements/graphics.txt", versions="loose"),
        "build": parse_requirements("requirements/build.txt", versions="loose"),
        "docs": parse_requirements("requirements/docs.txt", versions="loose"),
        "optional": parse_requirements("requirements/optional.txt", versions="loose"),
        "runtime": parse_requirements("requirements/runtime.txt", versions="loose"),
        "tests": parse_requirements("requirements/tests.txt", versions="loose"),
        "all-strict": parse_requirements("requirements.txt", versions="strict"),
        "headless-strict": parse_requirements(
            "requirements/headless.txt", versions="strict"
        ),
        "graphics-strict": parse_requirements(
            "requirements/graphics.txt", versions="strict"
        ),
        "build-strict": parse_requirements("requirements/build.txt", versions="strict"),
        "docs-strict": parse_requirements("requirements/docs.txt", versions="strict"),
        "optional-strict": parse_requirements(
            "requirements/optional.txt", versions="strict"
        ),
        "runtime-strict": parse_requirements(
            "requirements/runtime.txt", versions="strict"
        ),
        "tests-strict": parse_requirements("requirements/tests.txt", versions="strict"),
    }
    setupkw["name"] = NAME
    setupkw["version"] = VERSION
    setupkw["author"] = "Jon Crall"
    setupkw["author_email"] = "erotemic@gmail.com"
    setupkw["url"] = "https://github.com/Erotemic/guitool_ibeis"
    setupkw["description"] = "Guitool - tools for PyQt5 guis"
    setupkw["long_description"] = parse_description()
    setupkw["long_description_content_type"] = "text/x-rst"
    setupkw["license"] = "Apache 2"
    setupkw["packages"] = find_packages(".")
    setupkw["python_requires"] = ">=3.8"
    setupkw["classifiers"] = [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ]
    setup(**setupkw)



try:
    from guitool_ibeis import api_tree_node_cython
except ImportError:
    print('[test_treenode] cython OFF')
else:
    print('[test_treenode] cython ON')
from guitool_ibeis import api_tree_node


def test_build_internal_structure():
    """

    CommandLine:
        python -m guitool_ibeis.tests.test_treenode --test-test_build_internal_structure

    Example:
        >>> # DISABLE_DOCTEST
        >>> from guitool_ibeis.tests.test_treenode import *  # NOQA
        >>> test_build_internal_structure()
    """
    if 'api_tree_node_cython' in globals():
        print('test cython ON')
        _test_build_internal_structure(api_tree_node_cython, 'cython')
    else:
        print('test cython OFF')

    _test_build_internal_structure(api_tree_node, 'python')
    print('finished all tests')


def _test_build_internal_structure(_module, lang):
    import utool as ut
    # Test data
    N = 6
    #N = 2000

    def ider_level0():
        return range(N)

    def ider_level1(input_):
        _single = lambda x: [y for y in range(x ** 2, x ** 2 + max(0, ((N // 1) - x - 1)))]  # NOQA
        if isinstance(input_, list):
            return [_single(x) for x in input_]
        else:
            x = input_
            return _single(x)

    # Build Structure
    ider_list = [ider_level0, ider_level1]
    num_levels = len(ider_list)
    # TEST RECURSIVE
    print('================')
    with ut.Timer(lang + ' recursive:'):
        if num_levels == 0:
            root_id_list = []
        else:
            root_id_list = ider_list[0]()
        root_node1 = _module.TreeNode(-1, None, -1)
        level = 0
        _module._populate_tree_recursive(
            root_node1, root_id_list, num_levels, ider_list, level)
    if N < 10:
        print('')
        print(api_tree_node.tree_node_string(root_node1, indent=' *  '))
    print('================')
    #with ut.Timer(lang + ' iterative:'):
    #    # TEST ITERATIVE
    #    # TODO: Vet this code a bit more.
    #    root_node2 = _module.TreeNode(-1, None, -1)
    #    _module._populate_tree_iterative(
    #        root_node2, num_levels, ider_list)
    #if N < 10:
    #    print('')
    #    print(api_tree_node.tree_node_string(root_node2, indent=' *  '))
    print('================')
    print('finished %s test' % lang)


if __name__ == '__main__':
    """
    CommandLine:
        python -m guitool_ibeis.tests.test_treenode
        python -m guitool_ibeis.tests.test_treenode --allexamples
        python -m guitool_ibeis.tests.test_treenode --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



from __future__ import absolute_import, division, print_function
from guitool_ibeis.__PYQT__ import QtGui, QtCore  # NOQA
from guitool_ibeis.__PYQT__ import QtWidgets  # NOQA
from guitool_ibeis import guitool_components
import utool
#(print, print_, printDBG, rrr, profile) = utool.inject(__name__, '[APIButtonWidget]', DEBUG=False)
import utool as ut
ut.noinject(__name__, '[api_button_delegate]', DEBUG=False)


#DELEGATE_BASE = QtWidgets.QItemDelegate
DELEGATE_BASE = QtWidgets.QStyledItemDelegate


def rgb_to_qcolor(rgb):
    return QtGui.QColor(*rgb[0:3])


def rgb_to_qbrush(rgb):
    return QtGui.QBrush(rgb_to_qcolor(rgb))


def paint_button(painter, option, text='button', pressed=True, bgcolor=None,
                 fgcolor=None, clicked=None, button=None, view=None):
    #http://www.qtcentre.org/archive/index.php/t-31029.html
    opt = QtWidgets.QStyleOptionButton()
    opt.text = text
    opt.rect = option.rect
    opt.palette = option.palette
    if pressed:
        opt.state = QtWidgets.QStyle.State_Enabled | QtWidgets.QStyle.State_Sunken
    else:
        opt.state = QtWidgets.QStyle.State_Enabled | QtWidgets.QStyle.State_Raised

    #style = QtGui.Q Application.style()
    style = button.style()
    style.drawControl(QtWidgets.QStyle.CE_PushButton, opt, painter, button)


class APIButtonDelegate(DELEGATE_BASE):
    button_clicked = QtCore.pyqtSignal(QtCore.QModelIndex)

    def __init__(dgt, parent=None):
        assert parent is not None, 'parent must be a view'
        DELEGATE_BASE.__init__(dgt, parent)
        # FIXME: I don't like this state in the delegate, as it renders all
        # buttons
        dgt._pressed = None
        dgt.button_clicked.connect(dgt.on_button_click)

    def get_index_butkw(dgt, qtindex):
        """ The model data for a button should be a (text, callback) tuple.  OR
        it could be a function which accepts an qtindex and returns a button """
        data = qtindex.model().data(qtindex, QtCore.Qt.DisplayRole)
        # Get info
        if isinstance(data, tuple):
            buttontup = data
        elif utool.is_funclike(data):
            func = data
            buttontup = func(qtindex)
        else:
            raise AssertionError('bad type')
        text, callback = buttontup[0:2]
        butkw = {
            #'parent': dgt.parent(),
            'text': text,
            'clicked': callback,
        }
        if len(buttontup) > 2:
            butkw['bgcolor'] = buttontup[2]
            butkw['fgcolor'] = (0, 0, 0)
        return butkw

    def paint(dgt, painter, option, qtindex):
        painter.save()
        butkw = dgt.get_index_butkw(qtindex)
        # FIXME: I don't want to create a widget each time just
        # so I can access the button's style
        button = guitool_components.newButton(**butkw)
        pressed = dgt.is_qtindex_pressed(qtindex)
        view = dgt.parent()
        paint_button(painter, option, button=button, pressed=pressed,
                     view=view, **butkw)
        painter.restore()

    def is_qtindex_pressed(dgt, qtindex):
        return dgt._pressed is not None and dgt._pressed == (qtindex.row(), qtindex.column())

    @QtCore.pyqtSlot(QtCore.QModelIndex)
    def on_button_click(dgt, qtindex):
        if utool.VERBOSE:
            print('pressed button')
        butkw = dgt.get_index_butkw(qtindex)
        callback = butkw['clicked']
        callback()

    def editorEvent(dgt, event, model, option, qtindex):
        # http://stackoverflow.com/questions/14585575/button-delegate-issue
        #print('editor event')
        event_type = event.type()
        if event_type == QtCore.QEvent.MouseButtonPress:
            # store the position that is clicked
            dgt._pressed = (qtindex.row(), qtindex.column())
            if utool.VERBOSE:
                print('store')
            return True
        elif event_type == QtCore.QEvent.MouseButtonRelease:
            if dgt.is_qtindex_pressed(qtindex):
                print('emit')
                dgt.button_clicked.emit(qtindex)
                pass
            elif dgt._pressed is not None:
                # different place.
                # force a repaint on the pressed cell by emitting a dataChanged
                # Note: This is probably not the best idea
                # but I've yet to find a better solution.
                print('repaint')
                oldIndex = qtindex.model().index(*dgt._pressed)
                dgt._pressed = None
                qtindex.model().dataChanged.emit(oldIndex, oldIndex)
                pass
            dgt._pressed = None
            #print('mouse release')
            return True
        elif event_type == QtCore.QEvent.Leave:
            print('leave')
        elif event_type == QtCore.QEvent.MouseButtonDblClick:
            print('doubleclick')
        else:
            print('event_type = %r' % event_type)
            return DELEGATE_BASE.editorEvent(dgt, event, model, option, qtindex)


## graveyard:
#    #opt = QtWidgets.QStyleOptionViewItemV4(option)
#    #opt.initFrom(button)
#    #painter.drawRect(option.rect)
#    #print(style)
#    #if view is not None:
#    #view.style
#    ## FIXME: I cant set the colors!
#    #if bgcolor is not None:
#    #    opt.palette.setCurrentColorGroup(QtGui.QPalette.Normal)
#    #    opt.palette.setBrush(QtGui.QPalette.Normal, QtGui.QPalette.Button, rgb_to_qbrush(bgcolor))
#    #    opt.palette.setBrush(QtGui.QPalette.Base, rgb_to_qbrush(bgcolor))
#    #    opt.palette.setBrush(QtGui.QPalette.Window, rgb_to_qbrush(bgcolor))
#    #    opt.palette.setBrush(QtGui.QPalette.ButtonText, rgb_to_qbrush(bgcolor))
#    #    #
#    #    opt.palette.setColor(QtGui.QPalette.Normal, QtGui.QPalette.Button, rgb_to_qcolor(bgcolor))
#    #    opt.palette.setColor(QtGui.QPalette.Base, rgb_to_qcolor(bgcolor))
#    #    opt.palette.setColor(QtGui.QPalette.Window, rgb_to_qcolor(bgcolor))
#    #    opt.palette.setColor(QtGui.QPalette.ButtonText, rgb_to_qcolor(bgcolor))
#        #painter.setBrush(rgb_to_qbrush(bgcolor))
#    #if fgcolor is not None:
#    #    opt.palette.setBrush(QtGui.QPalette.Normal, QtGui.QPalette.ButtonText, rgb_to_qbrush(fgcolor))
#        #if not qtindex.isValid():
#        #    return None
#        #if view.indexWidget(qtindex):
#        #    return
#        #else:
#        #    view.setIndexWidget(qtindex, button)
#        #    # The view already has a button
#        #    # NOTE: this requires model::qtindex to be overwritten
#        #    # and return model.createIndex(row, col, object) where
#        #    # object is specified.
#        #    view.setIndexWidget(qtindex, None)
#        #    button = QtWidgets.QPushButton(text, view, clicked=view.cellButtonClicked)

#    #        pass
#    #    #       dgt._pressed = (qtindex.row(), qtindex.column())
#    #    #       return True
#    #        pass
#    #    else:
#    #        pass
#    #    #       if dgt._pressed == (qtindex.row(), qtindex.column()):
#    #    #           # we are at the same place, so emit
#    #    #           dgt.button_clicked.emit(*dgt._pressed)
#    #    #       elif dgt._pressed:
#    #    #           # different place.
#    #    #           # force a repaint on the pressed cell by emitting a dataChanged
#    #    #           # Note: This is probably not the best idea
#    #    #           # but I've yet to find a better solution.
#    #    #           oldIndex = qtindex.model().qtindex(*dgt._pressed)
#    #    #           dgt._pressed = None
#    #    #           qtindex.model().dataChanged.emit(oldIndex, oldIndex)
#    #    #       dgt._pressed = None
#    #    #       return True
#    #    #   else:
#    #    # default editor event;



# -*- coding: utf-8 -*-
"""
provides common methods for api_tree_view and api_table_view
"""
from __future__ import absolute_import, division, print_function
from guitool_ibeis.__PYQT__ import QtGui  # NOQA
from guitool_ibeis.__PYQT__ import QtCore
from guitool_ibeis.__PYQT__ import QtWidgets
from guitool_ibeis.__PYQT__.QtCore import Qt
import functools
from guitool_ibeis import qtype
from guitool_ibeis import api_button_delegate
from guitool_ibeis import api_thumb_delegate
from guitool_ibeis import guitool_main
from guitool_ibeis import guitool_misc
import utool
import utool as ut
import operator
from functools import reduce
# Valid API Models
from guitool_ibeis.stripe_proxy_model import StripeProxyModel
from guitool_ibeis.filter_proxy_model import FilterProxyModel
from guitool_ibeis.api_item_model import APIItemModel

(print, rrr, profile) = utool.inject2(__name__, '[APIItemView]')

VERBOSE_QT = ut.get_argflag(('--verbose-qt', '--verbqt'))
VERBOSE_ITEM_VIEW = ut.get_argflag(('--verbose-item-view'))
VERBOSE = utool.VERBOSE or VERBOSE_QT or VERBOSE_ITEM_VIEW

API_VIEW_BASE = QtWidgets.QAbstractItemView
ABSTRACT_VIEW_INJECT_KEY = ('QtWidgets.QAbstractItemView', 'guitool_ibeis')
register_view_method = utool.make_class_method_decorator(ABSTRACT_VIEW_INJECT_KEY, __name__)

injectviewinstance = functools.partial(utool.inject_instance, classkey=ABSTRACT_VIEW_INJECT_KEY)


VALID_API_MODELS = (FilterProxyModel, StripeProxyModel, APIItemModel)


class APIItemView(API_VIEW_BASE):
    """
    Trees and Tables implicitly inherit from this class.
    Abstractish class.

    other function in this file will be injected into the concrete
    implementations of either a table or tree view. The code is only written
    once but duplicated in each of the psuedo-children. It is done this way to
    avoid explicit multiple inheritance.
    """

    def __init__(view, parent=None):
        API_VIEW_BASE.__init__(view, parent)


@register_view_method
def _init_api_item_view(view):
    view.registered_single_keys = []
    view.registered_keypress_funcs = []

#---------------
# Data Manipulation
#---------------


@register_view_method
def _init_itemview_behavior(view):
    """

    Example:
        >>> # ENABLE_DOCTEST
        >>> # TODO figure out how to test these
        >>> # xdoctest: +REQUIRES(env:DISPLAY)
        >>> from guitool_ibeis.api_item_view import *  # NOQA
        >>> from guitool_ibeis import api_table_view
        >>> from guitool_ibeis import api_tree_view
        >>> import guitool_ibeis as gt
        >>> app = gt.ensure_qapp()[0]
        >>> view = api_table_view.APITableView()
        >>> view = api_tree_view.APITreeView()

    References:
        http://qt-project.org/doc/qt-4.8/qabstractitemview.html
    """
    # http://stackoverflow.com/questions/28680150/qtgui-qtableview-shows-data-in-background-while-a-cell-being-edited-pyqt4
    view.setAutoFillBackground(True)

    view.setWordWrap(True)

    # Selection behavior
    #view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
    #view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectColumns)
    view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectItems)

    # Selection behavior
    # view.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
    # view.setSelectionMode(QtWidgets.QAbstractItemView.ContiguousSelection)
    # view.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
    # view.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
    view.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

    # uniformRowHeights
    view.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerItem)
    view.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)

    # Allow sorting by column
    view.setSortingEnabled(True)

    # Edit Triggers
    #view.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)  # No Editing
    #view.setEditTriggers(QtWidgets.QAbstractItemView.SelectedClicked)
    #QtWidgets.QAbstractItemView.NoEditTriggers  |  # 0
    #QtWidgets.QAbstractItemView.CurrentChanged  |  # 1
    #QtWidgets.QAbstractItemView.DoubleClicked   |  # 2
    #QtWidgets.QAbstractItemView.SelectedClicked |  # 4
    #QtWidgets.QAbstractItemView.EditKeyPressed  |  # 8
    #QtWidgets.QAbstractItemView.AnyKeyPressed      # 16
    #view._defaultEditTriggers = QtWidgets.QAbstractItemView.AllEditTriggers

    bitwise_or = operator.__or__
    chosen_triggers = [
        #QtWidgets.QAbstractItemView.NoEditTriggers,
        QtWidgets.QAbstractItemView.CurrentChanged,
        QtWidgets.QAbstractItemView.DoubleClicked,
        QtWidgets.QAbstractItemView.SelectedClicked,
        QtWidgets.QAbstractItemView.EditKeyPressed,
        QtWidgets.QAbstractItemView.AnyKeyPressed,
    ]
    view._defaultEditTriggers = reduce(bitwise_or, chosen_triggers)
    #view._defaultEditTriggers = QtWidgets.QAbstractItemView.NoEditTriggers
    view.setEditTriggers(view._defaultEditTriggers)
    # TODO: Figure out how to not edit when you are selecting
    #view.setEditTriggers(QtWidgets.QAbstractItemView.AllEditTriggers)


@register_view_method
def infer_delegates(view, **headers):
    """ Infers which columns should be given item delegates """
    get_thumb_size = headers.get('get_thumb_size', None)
    col_type_list  = headers.get('col_type_list', [])
    num_cols = view.model().columnCount()
    num_duplicates = int(num_cols / len(col_type_list))
    col_type_list = col_type_list * num_duplicates
    view.has_thumbs = False
    for colx, coltype in enumerate(col_type_list):
        if coltype in  qtype.QT_PIXMAP_TYPES:
            if VERBOSE:
                print('[view] colx=%r is a PIXMAP' % colx)
            thumb_delegate = api_thumb_delegate.APIThumbDelegate(view, get_thumb_size)
            view.setItemDelegateForColumn(colx, thumb_delegate)
            view.has_thumbs = True
            # HACK
            #verticalHeader = view.verticalHeader()
            #verticalHeader.setDefaultSectionSize(256)
        elif coltype in qtype.QT_BUTTON_TYPES:
            if VERBOSE:
                print('[view] colx=%r is a BUTTON' % colx)
            button_delegate = api_button_delegate.APIButtonDelegate(view)
            view.setItemDelegateForColumn(colx, button_delegate)
        elif isinstance(coltype, QtWidgets.QAbstractItemDelegate):
            if VERBOSE:
                print('[view] colx=%r is a CUSTOM DELEGATE' % colx)
            view.setItemDelegateForColumn(colx, coltype)
        else:
            if VERBOSE:
                print('[view] colx=%r does not have a delgate' % colx)
            # Effectively unsets any existing delegates
            default_delegate = QtWidgets.QStyledItemDelegate(view)
            view.setItemDelegateForColumn(colx, default_delegate)


@register_view_method
def set_column_persistant_editor(view, column):
    """ Set each row in a column as persistant """
    num_rows = view.model.rowCount()
    print('view.set_persistant: %r rows' % num_rows)
    for row in range(num_rows):
        index  = view.model.index(row, column)
        view.view.openPersistentEditor(index)


@register_view_method
def _update_headers(view, **headers):
    """ Mirrors _update_headers in api_item_model """
    # Use headers from model #model = view.model #headers = model.headers
    # Get header info
    col_sort_index       = headers.get('col_sort_index', None)
    col_sort_reverse     = headers.get('col_sort_reverse', False)
    view.col_hidden_list = headers.get('col_hidden_list', [])
    view.col_name_list   = headers.get('col_name_list', [])
    # Call updates
    # FIXME: is this the right thing to do here?
    view._set_sort(col_sort_index, col_sort_reverse)
    view.infer_delegates(**headers)
    if ut.VERBOSE:
        print('[view] updating headers')
    col_width_list = headers.get('col_width_list', None)
    if col_width_list is not None:
        if isinstance(view, QtWidgets.QTreeView):
            horizontal_header = view.header()
        else:
            horizontal_header = view.horizontalHeader()
        for index, width in enumerate(col_width_list):
            # Check for functionally sepcified widths
            if hasattr(width, '__call__'):
                width = width()
            horizontal_header.resizeSection(index, width)
        #except AttributeError as ex:
        #    ut.embed()
        #    ut.printex(ex, 'tree view?')
        #view.infer_delegates_from_model(model=model) #view.resizeColumnsToContents()


@register_view_method
def _set_sort(view, col_sort_index, col_sort_reverse=True):
    if col_sort_index is not None:
        order = [Qt.AscendingOrder, Qt.DescendingOrder][col_sort_reverse]
        view.sortByColumn(col_sort_index, order)


@register_view_method
def hide_cols(view):
    total_num_cols = view.model().columnCount()
    num_cols = len(view.col_hidden_list)
    num_duplicates = int(total_num_cols / num_cols)
    duplicated_hidden_list = view.col_hidden_list * num_duplicates
    for col, hidden in enumerate(duplicated_hidden_list):
        view.setColumnHidden(col, hidden)


#@register_view_method
#def clear_selection(view):
#    #print('[api_item_view] clear_selection()')
#    selection_model = view.selectionModel()
#    selection_model.clearSelection()


@register_view_method
def get_row_and_qtindex_from_id(view, _id):
    """ uses an sqlrowid (from iders) to get a qtindex """
    model = view.model()
    qtindex, row = model.get_row_and_qtindex_from_id(_id)
    return qtindex, row


@register_view_method
def select_row_from_id(view, _id, scroll=False, collapse=True):
    """
        _id is from the iders function (i.e. an ibeis rowid)
        selects the row in that view if it exists
    """
    with ut.Timer('[api_item_view] select_row_from_id(id=%r, scroll=%r, collapse=%r)' %
                  (_id, scroll, collapse)):
        qtindex, row = view.get_row_and_qtindex_from_id(_id)
        if row is not None:
            if isinstance(view, QtWidgets.QTreeView):
                if collapse:
                    view.collapseAll()
                select_model = view.selectionModel()
                select_flag = QtCore.QItemSelectionModel.ClearAndSelect
                #select_flag = QtCore.QItemSelectionModel.Select
                #select_flag = QtCore.QItemSelectionModel.NoUpdate
                with ut.Timer('[api_item_view] selecting name. qtindex=%r' % (qtindex,)):
                    select_model.select(qtindex, select_flag)
                with ut.Timer('[api_item_view] expanding'):
                    view.setExpanded(qtindex, True)
            else:
                # For Table Views
                view.selectRow(row)
            # Scroll to selection
            if scroll:
                with ut.Timer('scrolling'):
                    view.scrollTo(qtindex)
            return row
    return None


@register_view_method
def connect_single_key_to_slot(view, key, func):
    """
    hacky way to simulate slots for generic key press events
    """
    view.registered_single_keys.append((key, func))


@register_view_method
def connect_keypress_to_slot(view, func):
    """
    hacky way to simulate slots for single key press events
    """
    view.registered_keypress_funcs.append(func)


@register_view_method
def selectedRows(view):
    selected_qtindex_list = view.selectedIndexes()
    selected_qtindex_list2 = []
    seen_ = set([])
    for qindex in selected_qtindex_list:
        row = qindex.row()
        if row not in seen_:
            selected_qtindex_list2.append(qindex)
            seen_.add(row)
    return selected_qtindex_list2


#---------------
# Qt Overrides
#---------------


def keyPressEvent(view, event):
    """
    Handles simple key press events. There is probably a better way to do this
    using real signals / slots, but maybe you need to always overwrite to set the
    handled flag correctly.

    CommandLine:
        xdoctest -m ~/code/guitool/guitool_ibeis/api_item_view.py keyPressEvent
        --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(env:DISPLAY)
        >>> from guitool_ibeis.api_item_view import *  # NOQA
        >>> import guitool_ibeis as gt
        >>> app = gt.ensure_qapp()[0]
        >>> wgt = gt.simple_api_item_widget()
        >>> view = wgt.view
        >>> c_pressed = [0]
        >>> def foo(view, event):
        >>>     key = event.key()
        >>>     print('[foo] Pressed key = %r' % (key,))
        >>>     if event.key() == Qt.Key_C:
        >>>         print('Pressed C')
        >>>         c_pressed[0] = 1
        >>>         return True
        >>> view.connect_keypress_to_slot(foo)
        >>> view._init_header_behavior()
        >>> # Try to simulate an event for testing
        >>> wgt.show()
        >>> from guitool_ibeis.__PYQT__ import QtTest, GUITOOL_PYQT_VERSION
        >>> QTest = QtTest.QTest
        >>> if GUITOOL_PYQT_VERSION == 4:
        >>>     QTest.qWaitForWindowShown(wgt)
        >>> else:
        >>>     QTest.qWaitForWindowActive(wgt)
        >>> qtindex = view.model().index(1, 2)
        >>> point = view.visualRect(qtindex).center()
        >>> #point = wgt.visibleRegion().boundingRect().center()
        >>> #QTest.mouseClick(view.viewport(), Qt.LeftButton, Qt.NoModifier, point)
        >>> QTest.mouseClick(wgt, Qt.LeftButton, Qt.NoModifier, point)
        >>> selected_indices = view.selectedIndexes()
        >>> print('selected_indices = %r' % (selected_indices,))
        >>> # Why does this not work?
        >>> def check_selection():
        >>>     selected_indices = view.selectedIndexes()
        >>>     if len(selected_indices) > 0:
        >>>         return selected_indices[0].data()
        >>> # Hack because I cant figure out how to get a click to simulate
        >>> # a selection
        >>> QTest.keyPress(view, Qt.Key_Right)
        >>> QTest.keyPress(view, Qt.Key_B)
        >>> ut.assert_eq(check_selection(), 'b')
        >>> QTest.keyPress(view, Qt.Key_C)
        >>> ut.assert_eq(check_selection(), 'b')
        >>> assert c_pressed[0] == 1
        >>> # xdoctest: +REQUIRES(--show)
        >>> ut.quit_if_noshow()
        >>> gt.qtapp_loop(wgt, frequency=100)
    """
    # TODO: can this be in api_item_view?
    assert isinstance(event, QtGui.QKeyEvent)
    if event.matches(QtGui.QKeySequence.Copy):
        #print('Received Ctrl+C in View')
        view.copy_selection_to_clipboard()
    #print ('[view] keyPressEvent: %s' % event.key())
    flag = False
    for func in view.registered_keypress_funcs:
        flag |= bool(func(view, event))
    for key, func in view.registered_single_keys:
        #print(key)
        if event.key() == key:
            flag = True
            func(view, event)
    if not flag:
        view.API_VIEW_BASE.keyPressEvent(view, event)


#@register_view_method
def itemDelegate(view, qindex):
    """ QtOverride: Returns item delegate for this index """
    # Does this even work? TODO: testme
    return API_VIEW_BASE.itemDelegate(view, qindex)


def setModel(view, model):
    """ QtOverride: Returns item delegate for this index """
    assert isinstance(model, VALID_API_MODELS), \
            ('APIItemViews only accepts APIItemModels (or one of its proxys),'
             'received a %r' % type(model))
    # Learn some things about the model before you fully connect it.
    if VERBOSE:
        print('[view] setting model')
    model._rows_updated.connect(view.on_rows_updated)
    #view.infer_delegates_from_model(model=model)
    # TODO: Update headers
    return view.API_VIEW_BASE.setModel(view, model)


#---------------
# Slots
#---------------


@register_view_method
def copy_selection_to_clipboard(view):
    """ Copys selected grid to clipboard """
    if VERBOSE:
        print('[guitool_ibeis] Copying selection to clipboard')
    copy_str = guitool_misc.get_view_selection_as_str(view)
    #copy_qstr = QtCore.Q__String(copy_str)
    copy_qstr = str(copy_str)
    clipboard = guitool_main.get_qtapp().clipboard()
    if VERBOSE:
        print(copy_str)
    clipboard.setText(copy_qstr)
    if VERBOSE:
        print('[guitool_ibeis] finished copy')


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m guitool_ibeis.api_item_view
        python -m guitool_ibeis.api_item_view --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



r"""
This module contains functions and clases to get data visualized fast (in
terms of development time)
"""
from __future__ import absolute_import, division, print_function
from guitool_ibeis.__PYQT__ import QtCore
from guitool_ibeis.__PYQT__ import QtWidgets
from guitool_ibeis.api_item_model import APIItemModel
from guitool_ibeis.api_table_view import APITableView
from guitool_ibeis.api_tree_view import APITreeView
from functools import partial
import utool as ut
import six
(print, rrr, profile) = ut.inject2(__name__, '[APIItemWidget]')


WIDGET_BASE = QtWidgets.QWidget

VERBOSE_ITEM_WIDGET = ut.get_argflag(('--verbose-item-widget', '--verbiw')) or ut.VERBOSE


def simple_api_item_widget():
    r"""
    Very simple example of basic APIItemWidget widget with CustomAPI

    CommandLine:
        python -m guitool_ibeis.api_item_widget --test-simple_api_item_widget
        python -m guitool_ibeis.api_item_widget --test-simple_api_item_widget --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(env:DISPLAY)
        >>> from guitool_ibeis.api_item_widget import *  # NOQA
        >>> import guitool_ibeis as gt
        >>> gt.ensure_qapp()  # must be ensured before any embeding
        >>> wgt = simple_api_item_widget()
        >>> # xdoctest: +REQUIRES(--show)
        >>> ut.quit_if_noshow()
        >>> wgt.show()
        >>> gt.qtapp_loop(wgt, frequency=100)
    """
    import guitool_ibeis as gt
    gt.ensure_qapp()
    col_getter_dict = {
        'col1': [1, 2, 3],
        'col2': ['a', 'b', 'c'],
        'col3': ['e', 'f', 'g'],
    }
    sortby = 'col1'

    #col_display_role_func_dict = {
    #    'col1': lambda x: 'banana %d' % ((x * 100 % 23)),
    #}

    api = gt.CustomAPI(
        col_getter_dict=col_getter_dict,
        #col_display_role_func_dict=col_display_role_func_dict,
        editable_colnames=['col3'],
        sortby=sortby)
    headers = api.make_headers(tblnice='Simple Example')

    wgt = gt.APIItemWidget()
    wgt.change_headers(headers)
    #gt.qtapp_loop(qwin=wgt, ipy=ipy, frequency=loop_freq)

    # for testing
    wgt.menubar = gt.newMenubar(wgt)
    wgt.menuFile = wgt.menubar.newMenu('Dev')
    def wgt_embed(wgt):
        view = wgt.view  # NOQA
        import utool
        utool.embed()
    ut.inject_func_as_method(wgt, wgt_embed)
    wgt.menuFile.newAction(triggered=wgt.wgt_embed)
    return wgt


def simple_api_tree_widget():
    r"""
    Minimal example of a tree based api widget

    CommandLine:
        python -m guitool_ibeis.api_item_widget --test-simple_api_tree_widget
        python -m guitool_ibeis.api_item_widget --test-simple_api_tree_widget --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(env:DISPLAY)
        >>> from guitool_ibeis.api_item_widget import *  # NOQA
        >>> import guitool_ibeis
        >>> guitool_ibeis.ensure_qapp()  # must be ensured before any embeding
        >>> wgt = simple_api_tree_widget()
        >>> # xdoctest: +REQUIRES(--show)
        >>> ut.quit_if_noshow()
        >>> wgt.show()
        >>> guitool_ibeis.qtapp_loop(wgt, frequency=100)
    """
    import guitool_ibeis
    guitool_ibeis.ensure_qapp()
    col_name_list = ['name', 'num_annots', 'annots']
    col_getter_dict = {
        'name': ['fred', 'sue', 'tom', 'mary', 'paul'],
        'num_annots': [2, 1, 3, 5, 1],
    }
    # make consistent data
    grouped_data = [
        [col_getter_dict['name'][index] + '-' + str(i) for i in range(num)]
        for index, num in enumerate(col_getter_dict['num_annots'])
    ]
    flat_data, reverse_list = ut.invertible_flatten1(grouped_data)
    col_getter_dict['annots'] = flat_data

    iders = [
        list(range(len(col_getter_dict['name']))),
        reverse_list
    ]

    col_level_dict = {
        'name': 0,
        'num_annots': 0,
        'annots': 1,
    }
    sortby = 'name'

    api = guitool_ibeis.CustomAPI(
        col_name_list=col_name_list,
        col_getter_dict=col_getter_dict,
        sortby=sortby,
        iders=iders,
        col_level_dict=col_level_dict,
    )
    headers = api.make_headers(tblnice='Tree Example')

    wgt = guitool_ibeis.APIItemWidget(view_class=APITreeView)
    wgt.change_headers(headers)
    #guitool_ibeis.qtapp_loop(qwin=wgt, ipy=ipy, frequency=loop_freq)
    return wgt


class CustomAPI(object):
    """
    Allows list of lists to be represented as an abstract api table

    # TODO: Rename CustomAPI
    API wrapper around a list of lists, each containing column data
    Defines a single table
    """
    def __init__(self, col_name_list=None, col_types_dict={}, col_getter_dict={},
                 col_bgrole_dict={}, col_ider_dict={}, col_setter_dict={},
                 editable_colnames={}, sortby=None, get_thumb_size=None,
                 sort_reverse=False, col_width_dict={}, strict=False,
                 col_display_role_func_dict=None, **kwargs):
        if VERBOSE_ITEM_WIDGET:
            print('[CustomAPI] <__init__>')
        if col_name_list is None:
            col_name_list = sorted(list(col_getter_dict.keys()))
        self.col_width_dict = col_width_dict
        self.col_name_list = []
        self.col_type_list = []
        self.col_getter_list = []
        self.col_setter_list = []
        self.col_display_role_func_dict = col_display_role_func_dict
        self.nCols = 0
        self.nRows = 0
        if get_thumb_size is None:
            self.get_thumb_size = lambda: 128
        else:
            self.get_thumb_size = get_thumb_size

        # Hack, maintain the original data
        # FIXME: make more ellegant
        self.orig_data_tup = (col_types_dict, col_getter_dict,
                              col_bgrole_dict, col_ider_dict, col_setter_dict,
                              editable_colnames, sortby, sort_reverse, strict)
        self.orig_kwargs = kwargs
        self.update_column_names(col_name_list)
        if VERBOSE_ITEM_WIDGET:
            print('[CustomAPI] </__init__>')

    def update_column_names(self, col_name_list):
        self.parse_column_tuples(col_name_list, *self.orig_data_tup, **self.orig_kwargs)

    def add_column_names(self, new_colnames):
        col_name_list = ut.unique(self.col_name_list + new_colnames)
        self.update_column_names(col_name_list)

    def get_available_colnames(self):
        col_getter_dict = self.orig_data_tup[1]
        return list(col_getter_dict.keys())

    def parse_column_tuples(self,
                            col_name_list,
                            col_types_dict,
                            col_getter_dict,
                            col_bgrole_dict,
                            col_ider_dict,
                            col_setter_dict,
                            editable_colnames,
                            sortby,
                            sort_reverse=True,
                            strict=False,
                            **kwargs):
        """
        parses simple lists into information suitable for making guitool_ibeis headers
        """
        # Unpack the column tuples into names, getters, and types
        if not strict:
            # slopply colname definitions
            flag_list = [colname in col_getter_dict for colname in col_name_list]
            if not all(flag_list):
                invalid_colnames = ut.compress(col_name_list, ut.not_list(flag_list))
                print('[api_item_widget] Warning: colnames=%r have no getters' % (
                    invalid_colnames,))
                col_name_list = ut.compress(col_name_list, flag_list)
            # sloppy type inference
            for colname in col_name_list:
                getter_ = col_getter_dict[colname]
                if colname not in col_types_dict:
                    type_ = ut.get_homogenous_list_type(getter_)
                    if type_ is not None:
                        col_types_dict[colname] = type_
        # sloppy kwargs.
        # FIXME: explicitly list col_nice_dict
        col_nice_dict = kwargs.get('col_nice_dict', {})
        self.col_nice_list = [col_nice_dict.get(name, name) for name in col_name_list]

        self.col_name_list = col_name_list
        self.col_type_list = [col_types_dict.get(colname, str)
                              for colname in col_name_list]
        # First col is always a getter
        self.col_getter_list = [col_getter_dict.get(colname, str)
                                for colname in col_name_list]
        # Get number of rows / columns
        self.nCols = len(self.col_getter_list)
        if self.nCols == 0:
            self.nRows = 0
        else:
            for getter in self.col_getter_list:
                if ut.isiterable(getter):
                    break
                getter = None
            # FIXME
            assert getter is not None, 'at least one getter must be an array/list'
            self.nRows = len(getter)

        #self.nRows = 0 if self.nCols == 0 else len(self.col_getter_list[0])  # FIXME
        # Init iders to default and then overwite based on dict inputs
        self.col_ider_list = [None] * self.nCols  # ut.alloc_nones(self.nCols)
        # for colname, ider_colnames in six.iteritems(col_ider_dict):
        # import utool
        # utool.embed()
        colname2_colx = ut.make_index_lookup(self.col_name_list)
        for colname, ider_colnames in six.iteritems(col_ider_dict):
            if colname not in colname2_colx:
                continue
            # for colname in self.col_name_list:
            ider_colnames = col_ider_dict[colname]
            try:
                colx = colname2_colx[colname]
                # Col iders might have tuple input
                ider_cols = self._uinput_1to1(self.col_name_list.index, ider_colnames)
                col_ider  = self._uinput_1to1(lambda c: ut.partial(self.get, c), ider_cols)
                self.col_ider_list[colx] = col_ider
                del col_ider
                del ider_cols
                del colx
                del colname
            except Exception as ex:
                ut.printex(ex, keys=['colname', 'ider_colnames', 'colx',
                                     'col_ider', 'ider_cols'])
                raise
        # Init setters to data, and then overwrite based on dict inputs
        self.col_setter_list = list(self.col_getter_list)
        for colname, col_setter in six.iteritems(col_setter_dict):
            colx = colname2_colx[colname]
            self.col_setter_list[colx] = col_setter
        # Init bgrole_getters to None, and then overwrite based on dict inputs
        self.col_bgrole_getter_list = [col_bgrole_dict.get(colname, None)
                                       for colname in self.col_name_list]
        # Mark edtiable columns
        self.col_edit_list = [name in editable_colnames for name in col_name_list]
        # Mark the sort column index
        if sortby is None:
            self.col_sort_index = 0
        elif ut.is_str(sortby):
            self.col_sort_index = self.col_name_list.index(sortby)
        else:
            self.col_sort_index = sortby
        self.col_sort_reverse = sort_reverse

        # Hacks for tree widget
        self._iders = kwargs.get('iders', None)
        col_level_dict = kwargs.get('col_level_dict', None)
        if col_level_dict is None:
            self.col_level_list = None
        else:
            self.col_level_list = ut.take(col_level_dict, col_name_list)

    def _infer_index(self, column, row):
        """
        returns the row based on the columns iders.
        This is the identity for the default ider
        """
        ider_ = self.col_ider_list[column]
        if ider_ is None:
            return row
        def iderfunc(func_):
            return func_(row)
        return self._uinput_1to1(iderfunc, ider_)

    @staticmethod
    def _uinput_1to1(func, input_):
        """ universal input (really just accept list or tuple as input to a list
        only function)

        Move to guitool_ibeis
        """
        if isinstance(input_, (tuple, list)):
            output_ = list(map(func, input_))
        else:
            output_ = func(input_)
        return output_

    @staticmethod
    def _general_get(getter, index, **kwargs):
        """ Works with getter funcs or indexable read/write arrays """
        if hasattr(getter, '__getitem__'):
            try:
                val = getter[index]
            except IndexError:
                print('index error at column index=%r' % (index))
                print('getter = %r' % (getter,))
                # print('index for column=%r' % (self.column_names[index]))
                raise
        else:
            val = getter(index, **kwargs)
        return val

    @staticmethod
    def _general_set(setter, index, val, **kwargs):
        """ Works with setter funcs or indexable read/write arrays """
        if hasattr(setter, '__setitem__'):
            setter[index] = val
        else:
            setter(index, val, **kwargs)

    def get(self, column, row, **kwargs):
        """
        getters always receive primary rowids, rectify if col_ider is
        specified (row might be a row_pair)
        """
        index = self._infer_index(column, row)
        column_getter = self.col_getter_list[column]
        # Columns might be getter funcs indexable read/write arrays
        try:
            return self._general_get(column_getter, index, **kwargs)
        except Exception:
            # FIXME: There may be an issue on tuple-key getters when row input is
            # vectorized. Hack it away
            if ut.isiterable(row):
                row_list = row
                return [self.get(column, row_, **kwargs) for row_ in row_list]
            else:
                raise

    def set(self, column, row, val):
        index = self._infer_index(column, row)
        column_setter = self.col_setter_list[column]
        # Columns might be setter funcs or indexable read/write arrays
        self._general_set(column_setter, index, val)

    def get_bgrole(self, column, row):
        bgrole_getter = self.col_bgrole_getter_list[column]
        if bgrole_getter is None:
            return None
        index = self._infer_index(column, row)
        return self._general_get(bgrole_getter, index)

    def ider(self):
        return list(range(self.nRows))

    def get_iders(self):
        if self._iders is None:
            return [self.ider]
        else:
            def _make_ider(level, ids_):
                # the first ider takes no args.
                # the second ider is in the context of the previous
                if level == 0:
                    def _tmp_ider():
                        return ids_
                else:
                    def _tmp_ider(parent_ids):
                        return ut.take(ids_, parent_ids)
                return _tmp_ider

            return [_make_ider(level, ids_) for level, ids_ in enumerate(self._iders)]

    def make_headers(self, tblname='custom_api', tblnice='Custom API'):
        """
        Builds headers for APIItemModel
        """
        headers = {
            'name': tblname,
            'nice': tblname if tblnice is None else tblnice,
            'iders': self.get_iders(),
            # 'iders': [self.ider],
            'col_name_list'    : self.col_name_list,
            'col_type_list'    : self.col_type_list,
            'col_nice_list'    : self.col_nice_list,
            'col_edit_list'    : self.col_edit_list,
            'col_sort_index'   : self.col_sort_index,
            'col_level_list'   : self.col_level_list,
            'col_sort_reverse' : self.col_sort_reverse,
            'col_getter_list'  : self._make_getter_list(),
            'col_setter_list'  : self._make_setter_list(),
            'col_bgrole_getter_list' : self._make_bgrole_getter_list(),
            'get_thumb_size'   : self.get_thumb_size,
            'col_display_role_func_dict': self.col_display_role_func_dict,
        }
        return headers

    def _make_bgrole_getter_list(self):
        return [partial(self.get_bgrole, column) for column in range(self.nCols)]

    def _make_getter_list(self):
        return [partial(self.get, column) for column in range(self.nCols)]

    def _make_setter_list(self):
        return [partial(self.set, column) for column in range(self.nCols)]


class APIItemWidget(WIDGET_BASE):
    """
    SIMPLE WIDGET WHICH AUTO-CREATES MODEL AND VIEW FOR YOU.
    """

    def __init__(widget, headers=None, parent=None,
                 model_class=APIItemModel, view_class=APITableView,
                 tblnice='APIItemWidget', doubleClicked=None):
        WIDGET_BASE.__init__(widget, parent)
        if isinstance(view_class, six.string_types):
            if view_class == 'tree':
                view_class = APITreeView
            elif view_class == 'table':
                view_class = APITableView
            else:
                raise ValueError('Unknown view_class=%r' % (view_class,))
        # Create vertical layout for the table to go into
        widget.vert_layout = QtWidgets.QVBoxLayout(widget)
        # Create a ColumnListTableView for the AbstractItemModel
        widget.view = view_class(parent=widget)
        # Instantiate the AbstractItemModel
        # FIXME: It is very bad to give the model a view.
        # Only the view should have a model
        # Also the model should not be an attribute because
        # it might live in another thread
        widget._model = model_class(parent=widget.view)

        widget.view.setModel(widget._model)
        widget.vert_layout.addWidget(widget.view)
        widget.tblnice = tblnice
        if headers is not None:
            # Make sure we don't call a subclass method
            APIItemWidget.change_headers(widget, headers)
        widget.connect_signals()
        widget.api = None
        if doubleClicked:
            widget.view.doubleClicked.connect(doubleClicked)

    @property
    def model(widget):
        return widget.view.model()

    def connect_api(widget, api, autopopulate=True):
        widget.api = api
        if autopopulate:
            widget.refresh_headers()
            #widget.resize_headers()
            #headers = api.make_headers(tblnice=widget.tblnice)
            #widget.change_headers(headers)
            #print(ut.repr2(headers))

    def change_headers(widget, headers):
        parent = widget.parent()
        # Update headers of both model and view
        widget._model._update_headers(**headers)
        widget.view._update_headers(**headers)
        if parent is None:
            nice = headers.get('nice', 'NO NICE NAME')
            widget.setWindowTitle(nice)

    def connect_signals(widget):
        widget._model._rows_updated.connect(widget.on_rows_updated)
        widget.view.contextMenuClicked.connect(widget.on_contextMenuRequested)

    def on_rows_updated(widget, name, num):
        if VERBOSE_ITEM_WIDGET:
            print('rows updated')
        pass

    def refresh_headers(widget):
        headers = widget.api.make_headers(tblnice=widget.tblnice)
        widget.change_headers(headers)
        #print(ut.repr2(headers))

    def resize_headers(self, api=None):
        if api is None:
            api = self.api
        horizontal_header = self.view.horizontalHeader()
        for col, width in six.iteritems(api.col_width_dict):
            #horizontal_header.defaultSectionSize()
            try:
                index = api.col_name_list.index(col)
            except ValueError:
                pass
            else:
                horizontal_header.resizeSection(index, width)

    @QtCore.pyqtSlot(QtCore.QModelIndex, QtCore.QPoint)
    def on_contextMenuRequested(widget, index, pos):
        print('context request')
        if widget.api is not None:
            print(ut.repr2(widget.api.get_available_colnames()))
            # HACK test
            #widget.api.add_column_names(['qx2_gt_rank', 'qx2_gf_rank', 'qx2_gt_raw_score', 'qx2_gf_raw_score'])
            widget.refresh_headers()
            #widget.change_headers(widget.api.make_headers())
        if VERBOSE_ITEM_WIDGET:
            print('context request')


if __name__ == '__main__':
    """
    CommandLine:
        python -m guitool_ibeis.api_item_widget
        python -m guitool_ibeis.api_item_widget --allexamples
        python -m guitool_ibeis.api_item_widget --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
from guitool_ibeis.__PYQT__ import QtCore
from guitool_ibeis.__PYQT__ import GUITOOL_PYQT_VERSION
from guitool_ibeis.__PYQT__ import QtWidgets
from guitool_ibeis.guitool_decorators import signal_, slot_
import utool as ut
from guitool_ibeis import api_item_view

(print, rrr, profile) = ut.inject2(__name__)


# If you need to set the selected index try:
# AbstractItemView::setCurrentIndex
# AbstractItemView::scrollTo
# AbstractItemView::keyboardSearch

API_VIEW_BASE = QtWidgets.QTreeView
#API_VIEW_BASE = QtWidgets.QAbstractItemView


class APITreeView(API_VIEW_BASE):
    """
    Tree view of API data.
    Implicitly inherits from APIItemView
    """
    rows_updated = signal_(str, int)
    contextMenuClicked = signal_(QtCore.QModelIndex, QtCore.QPoint)
    API_VIEW_BASE = API_VIEW_BASE

    def __init__(view, parent=None):
        # Qt Inheritance
        API_VIEW_BASE.__init__(view, parent)
        # Implicitly inject common APIItemView functions
        api_item_view.injectviewinstance(view)
        view._init_itemview_behavior()
        view._init_tree_behavior()
        view.col_hidden_list = []
        ##view._init_header_behavior()
        # Context menu
        view.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        view.customContextMenuRequested.connect(view.on_customMenuRequested)
        #view.cornerButton = guitool_components.newButton(view)
        #view.setCornerWidget(view.cornerButton)
        view._init_api_item_view()

        #view.setUniformRowHeights(True)

    #---------------
    # Initialization
    #---------------

    def _init_tree_behavior(view):
        """ Tree behavior

        SeeAlso:
            api_item_view._init_itemview_behavior
        """
        pass

    def _init_header_behavior(view):
        """ Header behavior

        CommandLine:
            python -m guitool_ibeis.api_tree_view --test-_init_header_behavior

        Example:
            >>> # ENABLE_DOCTEST
            >>> # TODO figure out how to test these
            >>> # xdoctest: +REQUIRES(env:DISPLAY)
            >>> from guitool_ibeis.api_tree_view import *  # NOQA
            >>> import guitool_ibeis as gt
            >>> app = gt.ensure_qapp()
            >>> view = APITreeView()
            >>> view._init_header_behavior()
        """
        # Row Headers
        # Column headers
        header = view.header()
        header.setVisible(True)
        header.setStretchLastSection(True)
        header.setSortIndicatorShown(True)
        header.setHighlightSections(True)
        # Column Sizes
        # DO NOT USE RESIZETOCONTENTS. IT MAKES THINGS VERY SLOW
        #horizontalHeader.setResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        #horizontalHeader.setResizeMode(QtWidgets.QHeaderView.Stretch)
        if GUITOOL_PYQT_VERSION == 4:
            header.setResizeMode(QtWidgets.QHeaderView.Interactive)
            #horizontalHeader.setCascadingSectionResizes(True)
            # Columns moveable
            header.setMovable(True)
        else:
            header.setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
            header.setSectionsMovable(True)

    #---------------
    # Qt Overrides
    #---------------

    def setModel(view, model):
        """ QtOverride: Returns item delegate for this index """
        api_item_view.setModel(view, model)

    def keyPressEvent(view, event):
        return api_item_view.keyPressEvent(view, event)

    #---------------
    # Slots
    #---------------

    @slot_(str, int)
    def on_rows_updated(view, tblname, num):
        # re-emit the model signal
        view.rows_updated.emit(tblname, num)

    @slot_(QtCore.QPoint)
    def on_customMenuRequested(view, pos):
        index = view.indexAt(pos)
        view.contextMenuClicked.emit(index, pos)


def testdata_tree_view():
    r"""
    CommandLine:
        python -m guitool_ibeis.api_tree_view testdata_tree_view
        python -m guitool_ibeis.api_tree_view testdata_tree_view --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> import guitool_ibeis as gt
        >>> from guitool_ibeis.api_tree_view import *  # NOQA
        >>> wgt = testdata_tree_view()
        >>> view = wgt.view
        >>> rows = view.selectedRows()
        >>> print('rows = %r' % (rows,))
        >>> # xdoctest: +REQUIRES(--show)
        >>> ut.quit_if_noshow()
        >>> gt.qtapp_loop(qwin=wgt)
    """
    import guitool_ibeis as gt
    gt.ensure_qapp()
    col_name_list = ['name', 'num_annots', 'annots']
    col_getter_dict = {
        'name': ['fred', 'sue', 'tom', 'mary', 'paul'],
        'num_annots': [2, 1, 3, 5, 1],
    }
    # make consistent data
    grouped_data = [
        [col_getter_dict['name'][index] + '-' + str(i) for i in range(num)]
        for index, num in enumerate(col_getter_dict['num_annots'])
    ]
    flat_data, reverse_list = ut.invertible_flatten1(grouped_data)
    col_getter_dict['annots'] = flat_data

    iders = [
        list(range(len(col_getter_dict['name']))),
        reverse_list
    ]

    col_level_dict = {
        'name': 0,
        'num_annots': 0,
        'annots': 1,
    }
    sortby = 'name'

    api = gt.CustomAPI(
        col_name_list=col_name_list,
        col_getter_dict=col_getter_dict,
        sortby=sortby,
        iders=iders,
        col_level_dict=col_level_dict,
    )
    headers = api.make_headers(tblnice='Tree Example')

    wgt = gt.APIItemWidget(view_class=APITreeView)
    wgt.change_headers(headers)

    wgt.menubar = gt.newMenubar(wgt)
    wgt.menuFile = wgt.menubar.newMenu('Dev')

    def wgt_embed(wgt):
        view = wgt.view  # NOQA
        import utool
        utool.embed()
    ut.inject_func_as_method(wgt, wgt_embed)
    wgt.menuFile.newAction(triggered=wgt.wgt_embed)

    return wgt


if __name__ == '__main__':
    """
    CommandLine:
        python -m guitool_ibeis.api_tree_view
        python -m guitool_ibeis.api_tree_view --allexamples
        python -m guitool_ibeis.api_tree_view --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



"""
CommandLine:
    rm -rf /media/raid/work/PZ_MTEST/_ibsdb/_ibeis_cache/match_thumbs/
    python -m ibeis.gui.inspect_gui --test-test_review_widget --show --verbose-thumb
"""
from __future__ import absolute_import, division, print_function
from guitool_ibeis.__PYQT__ import QtGui, QtCore
from guitool_ibeis.__PYQT__ import QtWidgets  # NOQA
import six
from os.path import exists
import utool as ut
ut.noinject(__name__, '[APIThumbDelegate]')


VERBOSE_QT = ut.get_argflag(('--verbose-qt', '--verbqt'))
VERBOSE_THUMB = ut.VERBOSE or ut.get_argflag(('--verbose-thumb', '--verbthumb')) or VERBOSE_QT


MAX_NUM_THUMB_THREADS = 1


def read_thumb_size(thumb_path):
    import vtool_ibeis as vt
    if VERBOSE_THUMB:
        print('[ThumbDelegate] Reading thumb size')
    # npimg = vt.imread(thumb_path, delete_if_corrupted=True)
    # (height, width) = npimg.shape[0:2]
    # del npimg
    try:
        width, height = vt.open_image_size(thumb_path)
    except IOError as ex:
        if ut.checkpath(thumb_path, verbose=True):
            ut.printex(ex, 'image=%r seems corrupted. Needs deletion' %
                       (thumb_path,), iswarning=True)
            ut.delete(thumb_path)
        else:
            ut.printex(ex, 'image=%r does not exist', (thumb_path,),
                       iswarning=True)
        raise
    return width, height


def test_show_qimg(qimg):
    qpixmap = QtGui.QPixmap(qimg)
    lbl = QtWidgets.QLabel()
    lbl.setPixmap(qpixmap)
    lbl.show()   # show label with qim image
    return lbl


#@ut.memprof
def read_thumb_as_qimg(thumb_path):
    r"""
    Args:
        thumb_path (?):

    Returns:
        tuple: (qimg, width, height)

    CommandLine:
        python -m guitool_ibeis.api_thumb_delegate --test-read_thumb_as_qimg --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(env:DISPLAY)
        >>> from guitool_ibeis.api_thumb_delegate import *  # NOQA
        >>> import guitool_ibeis
        >>> # build test data
        >>> thumb_path = ut.grab_test_imgpath('carl.jpg')
        >>> # execute function
        >>> guitool_ibeis.ensure_qtapp()
        >>> qimg = read_thumb_as_qimg(thumb_path)
        >>> print(qimg)
        >>> # xdoctest: +REQUIRES(--show)
        >>> lbl = test_show_qimg(qimg)
        >>> #guitool_ibeis.qtapp_loop()
        >>> # verify results

    Timeit::
        %timeit np.dstack((npimg, np.full(npimg.shape[0:2], 255, dtype=np.uint8)))
        %timeit cv2.cvtColor(npimg, cv2.COLOR_BGR2BGRA)
        npimg1 = np.dstack((npimg, np.full(npimg.shape[0:2], 255, dtype=np.uint8)))
        # seems to be memory leak in cvtColor?
        npimg2 = cv2.cvtColor(npimg, cv2.COLOR_BGR2BGRA)

    """
    if VERBOSE_THUMB:
        print('[ThumbDelegate] Reading thumb as qimg. thumb_path=%r' % (thumb_path,))
    # Read thumbnail image and convert to 32bit aligned for Qt
    #if False:
    #    data  = np.dstack((npimg, np.full(npimg.shape[0:2], 255, dtype=np.uint8)))
    #if False:
    #    # Reading the npimage and then handing it off to Qt causes a memory
    #    # leak. The numpy array probably is never unallocated because qt doesn't
    #    # own it and it never loses its reference count
    #    #npimg = vt.imread(thumb_path, delete_if_corrupted=True)
    #    #print('npimg.dtype = %r, %r' % (npimg.shape, npimg.dtype))
    #    #npimg   = cv2.cvtColor(npimg, cv2.COLOR_BGR2BGRA)
    #    #format_ = QtGui.QImage.Format_ARGB32
    #    ##    #data    = npimg.astype(np.uint8)
    #    ##    #npimg   = np.dstack((npimg[:, :, 3], npimg[:, :, 0:2]))
    #    ##    #data    = npimg.astype(np.uint8)
    #    ##else:
    #    ## Memory seems to be no freed by the QImage?
    #    ##data = np.ascontiguousarray(npimg[:, :, ::-1].astype(np.uint8), dtype=np.uint8)
    #    ##data = np.ascontiguousarray(npimg[:, :, :].astype(np.uint8), dtype=np.uint8)
    #    #data = npimg
    #    ##format_ = QtGui.QImage.Format_RGB888
    #    #(height, width) = data.shape[0:2]
    #    #qimg    = QtGui.QImage(data, width, height, format_)
    #    #del npimg
    #    #del data
    #else:
    #format_ = QtGui.QImage.Format_ARGB32
    #qimg    = QtGui.QImage(thumb_path, format_)
    qimg    = QtGui.QImage(thumb_path)
    return qimg


RUNNING_CREATION_THREADS = {}


def register_thread(key, val):
    global RUNNING_CREATION_THREADS
    RUNNING_CREATION_THREADS[key] = val


def unregister_thread(key):
    global RUNNING_CREATION_THREADS
    del RUNNING_CREATION_THREADS[key]


DELEGATE_BASE = QtWidgets.QItemDelegate


class APIThumbDelegate(DELEGATE_BASE):
    """
    There is one Thumb Delegate per column. Keep that in mind when writing for
    this class.

    TODO: The delegate can have a reference to the view, and it is allowed
    to resize the rows to fit the images.  It probably should not resize columns
    but it can get the column width and resize the image to that size.

    get_thumb_size is a callback function which should return whatever the
    requested thumbnail size is

    SeeAlso:
         api_item_view.infer_delegates
    """
    def __init__(dgt, parent=None, get_thumb_size=None):
        if VERBOSE_THUMB:
            print('[ThumbDelegate] __init__ parent=%r, get_thumb_size=%r' %
                    (parent, get_thumb_size))
        DELEGATE_BASE.__init__(dgt, parent)
        dgt.pool = None
        # TODO: get from the view
        if get_thumb_size is None:
            dgt.get_thumb_size = lambda: 128  # 256
        else:
            dgt.get_thumb_size = get_thumb_size  # 256
        dgt.last_thumbsize = None
        dgt.row_rezised_flags = {}  # SUPER HACK FOR RESIZE SHRINK
        try:
            import cachetools
            dgt.thumb_cache = cachetools.TTLCache(256, ttl=2)
        except ImportError:
            dgt.thumb_cache = ut.LRUDict(256)
        #import utool
        #utool.embed()

    def paint(dgt, painter, option, qtindex):
        """
        TODO: prevent recursive paint
        """
        view = dgt.parent()
        offset = view.verticalOffset() + option.rect.y()
        # Check if still in viewport
        if view_would_not_be_visible(view, offset):
            return None
        try:
            thumb_path = dgt.get_thumb_path_if_exists(view, offset, qtindex)
            if thumb_path is not None:
                # Check if still in viewport
                if view_would_not_be_visible(view, offset):
                    return None
                # Read the precomputed thumbnail
                if thumb_path in dgt.thumb_cache:
                    qimg = dgt.thumb_cache[thumb_path]
                else:
                    qimg = read_thumb_as_qimg(thumb_path)
                    dgt.thumb_cache[thumb_path] = qimg
                width, height = qimg.width(), qimg.height()
                # Adjust the cell size to fit the image
                dgt.adjust_thumb_cell_size(qtindex, width, height)
                # Check if still in viewport
                if view_would_not_be_visible(view, offset):
                    return None
                # Paint image on an item in some view
                painter.save()
                painter.setClipRect(option.rect)
                painter.translate(option.rect.x(), option.rect.y())
                painter.drawImage(QtCore.QRectF(0, 0, width, height), qimg)
                painter.restore()
        except Exception as ex:
            print('Error in APIThumbDelegate')
            ut.printex(ex, 'Error in APIThumbDelegate', tb=True)
            painter.save()
            painter.restore()

    def sizeHint(dgt, option, qtindex):
        view = dgt.parent()
        offset = view.verticalOffset() + option.rect.y()
        try:
            thumb_path = dgt.get_thumb_path_if_exists(view, offset, qtindex)
            if thumb_path is not None:
                # Read the precomputed thumbnail
                width, height = read_thumb_size(thumb_path)
                return QtCore.QSize(width, height)
            else:
                #print("[APIThumbDelegate] Name not found")
                return QtCore.QSize()
        except Exception as ex:
            print('Error in APIThumbDelegate')
            ut.printex(ex, 'Error in APIThumbDelegate', tb=True,
                       iswarning=True)
            return QtCore.QSize()

    def get_model_data(dgt, qtindex):
        """
        The model data for a thumb should be a tuple:
        (thumb_path, img_path, imgsize, bboxes, thetas)
        """
        model = qtindex.model()
        datakw = dict(thumbsize=dgt.get_thumb_size())
        data = model.data(qtindex, QtCore.Qt.DisplayRole, **datakw)
        if data is None:
            return None
        # The data should be specified as a thumbtup
        #if isinstance(data, QtCore.QVariant):
        if hasattr(data, 'toPyObject'):
            data = data.toPyObject()
        if data is None:
            return None
        if isinstance(data, six.string_types):
            # data = (data, None, None, None, None)
            return data
        if isinstance(data, dict):
            # HACK FOR DIFFERENT TYPE OF THUMB DATA
            return data
        assert isinstance(data, tuple), (
            'data=%r is %r. should be a thumbtup' % (data, type(data)))
        thumbtup = data
        #(thumb_path, img_path, bbox_list) = thumbtup
        return thumbtup

    def spawn_thumb_creation_thread(dgt, thumb_path, img_path, img_size,
                                    qtindex, view, offset, bbox_list,
                                    theta_list, interest_list):
        if VERBOSE_THUMB:
            print('[ThumbDelegate] Spawning thumbnail creation thread')
        thumbsize = dgt.get_thumb_size()
        thumb_creation_thread = ThumbnailCreationThread(
            thumb_path, img_path, img_size, thumbsize,
            qtindex, view, offset, bbox_list, theta_list,
            interest_list
        )
        #register_thread(thumb_path, thumb_creation_thread)
        # Initialize threadcount
        if dgt.pool is None:
            #dgt.pool = QtCore.QThreadPool()
            #dgt.pool.setMaxThreadCount(MAX_NUM_THUMB_THREADS)
            dgt.pool = QtCore.QThreadPool.globalInstance()
        dgt.pool.start(thumb_creation_thread)
        # print('[ThumbDelegate] Waiting to compute')

    def get_thumb_path_if_exists(dgt, view, offset, qtindex):
        """
        Checks if the thumbnail is ready to paint

        Returns:
            thumb_path if computed otherwise returns None
        """

        # Check if still in viewport
        if view_would_not_be_visible(view, offset):
            return None

        # Get data from the models display role
        try:
            data = dgt.get_model_data(qtindex)
            if data is None:
                if VERBOSE_THUMB:
                    print('[thumb_delegate] no data')
                return
            thumbtup_mode = isinstance(data, tuple)
            thumbdat_mode = isinstance(data, dict)
            if isinstance(data, six.string_types):
                thumb_path = data
                assert exists(thumb_path), 'must exist'
                return thumb_path
            if thumbtup_mode:
                if len(data) == 5:
                    (thumb_path, img_path, img_size, bbox_list, theta_list) = data
                    interest_list = []
                else:
                    (thumb_path, img_path, img_size, bbox_list, theta_list,
                     interest_list) = data
                invalid = (thumb_path is None or img_path is None or bbox_list is
                           None or img_size is None)
                if invalid:
                    print('[thumb_delegate] something is wrong')
                    return
            elif thumbdat_mode:
                thumb_path = data['fpath']
            else:
                print('[thumb_delegate] something is wrong')
                return
        except AssertionError as ex:
            ut.printex(ex, 'error getting thumbnail data')
            return

        # Check if still in viewport
        if view_would_not_be_visible(view, offset):
            return None

        if not exists(thumb_path):
            if thumbtup_mode:
                if not exists(img_path):
                    if VERBOSE_THUMB:
                        print('[ThumbDelegate] SOURCE IMAGE NOT COMPUTED: %r' %
                              (img_path,))
                    return None
                dgt.spawn_thumb_creation_thread(
                    thumb_path, img_path, img_size, qtindex, view, offset,
                    bbox_list, theta_list, interest_list)
                return None
            elif thumbdat_mode:
                thumbdat = data
                thread_func = thumbdat['thread_func']
                main_func = thumbdat['main_func']
                #kwargs = data['kwargs']
                #func(*args, **kwargs)
                #print('data = %r' % (data,))
                #print('newdata not computed')
                # SPAWN
                if VERBOSE_THUMB:
                    print('[ThumbDelegate] Spawning thumbnail creation thread')
                args = main_func()
                thumb_creation_thread = ThumbnailCreationThread2(
                    thread_func, args, qtindex, view, offset
                )
                #register_thread(thumb_path, thumb_creation_thread)
                # Initialize threadcount
                if dgt.pool is None:
                    #dgt.pool = QtCore.QThreadPool()
                    #dgt.pool.setMaxThreadCount(MAX_NUM_THUMB_THREADS)
                    dgt.pool = QtCore.QThreadPool.globalInstance()
                dgt.pool.start(thumb_creation_thread)
                # print('[ThumbDelegate] Waiting to compute')
                return None
        else:
            # thumb is computed return the path
            return thumb_path

    def adjust_thumb_cell_size(dgt, qtindex, width, height):
        """
        called during paint to ensure that the cell is large enough for the
        image.
        """
        view = dgt.parent()
        if isinstance(view, QtWidgets.QTableView):
            # dimensions of the table cells
            row = qtindex.row()
            col_width = view.columnWidth(qtindex.column())
            col_height = view.rowHeight(row)
            thumbsize = dgt.get_thumb_size()
            if thumbsize != dgt.last_thumbsize:
                # has thumbsize changed?
                if thumbsize != col_width:
                    view.setColumnWidth(qtindex.column(), thumbsize)
                if height != col_height:
                    view.setRowHeight(qtindex.row(), height)
                dgt.last_thumbsize = thumbsize
            # Let columns shrink
            if thumbsize != col_width:
                view.setColumnWidth(qtindex.column(), thumbsize)
            # Let rows grow
            if height > col_height:
                view.setRowHeight(qtindex.row(), height)
            if dgt.row_rezised_flags.get(row):
                # HACK TO ONLY SHRINK ONCE WONT WORK WITH RESORT
                return
            else:
                dgt.row_rezised_flags[row] = True
                # Let rows shrink
                # IF THERE IS MORE THAN ONE COLUMN WITH THUMBS THEN THIS WILL CAUSE
                # COLS TO BE RESIZED MANY TIMES UNDER THE HOOD. THAT CAUSES
                # MULTIPLE READS OF THE THUMBS WHICH CAUSES MAJOR SLOWDOWNS.
                if height < col_height:
                    view.setRowHeight(qtindex.row(), height)
        elif isinstance(view, QtWidgets.QTreeView):
            col_width = view.columnWidth(qtindex.column())
            col_height = view.rowHeight(qtindex)
            # TODO: finishme


def view_would_not_be_visible(view, offset):
    """
    Check if the current scroll position is far beyond the
    scroll position when this was initially requested.
    """
    viewport = view.viewport()
    height = viewport.size().height()
    height_offset = view.verticalOffset()
    current_offset = height_offset + height // 2
    return abs(current_offset - offset) >= height


def get_thread_thumb_info(bbox_list, theta_list, thumbsize, img_size):
    r"""
    CommandLine:
        python -m guitool_ibeis.api_thumb_delegate --test-get_thread_thumb_info

    Example:
        >>> # ENABLE_DOCTEST
        >>> from guitool_ibeis.api_thumb_delegate import *  # NOQA
        >>> # build test data
        >>> bbox_list = [(100, 50, 400, 200)]
        >>> theta_list = [0]
        >>> thumbsize = 128
        >>> img_size = 600, 300
        >>> # execute function
        >>> result = get_thread_thumb_info(bbox_list, theta_list, thumbsize, img_size)
        >>> # verify results
        >>> print(result)
        ((128, 64), [[[21, 11], [107, 11], [107, 53], [21, 53], [21, 11]]])

    """
    import vtool_ibeis as vt
    theta_list = [theta_list] if not ut.is_listlike(theta_list) else theta_list
    max_dsize = (thumbsize, thumbsize)
    dsize, sx, sy = vt.resized_clamped_thumb_dims(img_size, max_dsize)
    # Compute new verts list
    new_verts_list = list(vt.scaled_verts_from_bbox_gen(bbox_list, theta_list, sx, sy))
    return dsize, new_verts_list


def make_thread_thumb(img_path, dsize, new_verts_list, interest_list):
    r"""
    Makes thumbnail with overlay. Called in thead

    CommandLine:
        python -m guitool_ibeis.api_thumb_delegate --test-make_thread_thumb --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from guitool_ibeis.api_thumb_delegate import *  # NOQA
        >>> import plottool_ibeis as pt
        >>> # build test data
        >>> img_path = ut.grab_test_imgpath('carl.jpg')
        >>> dsize = (32, 32)
        >>> new_verts_list = []
        >>> # execute function
        >>> thumb = make_thread_thumb(img_path, dsize, new_verts_list)
        >>> ut.quit_if_noshow()
        >>> pt.imshow(thumb)
        >>> pt.show_if_requested()
    """
    import vtool_ibeis as vt
    from vtool_ibeis import geometry
    orange_bgr = (0, 128, 255)
    blue_bgr = (255, 128, 0)
    # imread causes a MEMORY LEAK most likely!
    img = vt.imread(img_path)  # Read Image (.0424s) <- Takes most time!
    #if False:
    #    #http://stackoverflow.com/questions/9794019/convert-numpy-array-to-pyside-qpixmap
    #    # http://kogs-www.informatik.uni-hamburg.de/~meine/software/vigraqt/qimage2ndarray.py
    #    #import numpy as np
    #    #qimg = QtGui.QImage(img_path, str(QtGui.QImage.Format_RGB32))
    #    #temp_shape = (qimg.height(), qimg.bytesPerLine() * 8 // qimg.depth(), 4)
    #    #result_shape = (qimg.height(), qimg.width())
    #    #buf = qimg.bits().asstring(qimg.numBytes())
    #    #result  = np.frombuffer(buf, np.uint8).reshape(temp_shape)
    #    #result = result[:, :result_shape[1]]
    #    #result = result[..., :3]
    #    #img = result
    thumb = vt.image.resize(img, dsize)  # Resize to thumb dims (.0015s)
    del img
    # Draw bboxes on thumb (not image)
    color_bgr_list = [blue_bgr if interest else orange_bgr for interest in interest_list]
    for new_verts, color_bgr in zip(new_verts_list, color_bgr_list):
        if new_verts is not None:
            geometry.draw_verts(thumb, new_verts, color=color_bgr, thickness=2, out=thumb)
        #thumb = geometry.draw_verts(thumb, new_verts, color=orange_bgr, thickness=2)
    return thumb


RUNNABLE_BASE = QtCore.QRunnable


class ThumbnailCreationThread2(RUNNABLE_BASE):
    """
    HACK
    TODO: http://stackoverflow.com/questions/6783194/background-thread-with-qthread-in-pyqt
    """

    def __init__(thread, thread_func, args, qtindex, view, offset):
        RUNNABLE_BASE.__init__(thread)
        thread.thread_func = thread_func
        thread.args = args
        thread.qtindex = qtindex
        thread.offset = offset
        thread.view = view

    def thumb_would_not_be_visible(thread):
        return view_would_not_be_visible(thread.view, thread.offset)

    def _run(thread):
        """ Compute thumbnail in a different thread """
        if thread.thumb_would_not_be_visible():
            return
        #func = thread.thumbdat['func']
        thread.thread_func(thread.thumb_would_not_be_visible, *thread.args)
        #func(check_func=thread.thumb_would_not_be_visible)
        thread.qtindex.model().dataChanged.emit(thread.qtindex, thread.qtindex)

    def run(thread):
        try:
            thread._run()
        except Exception as ex:
            ut.printex(ex, 'thread failed', tb=True)
            #raise


class ThumbnailCreationThread(RUNNABLE_BASE):
    """
    Helper to compute thumbnails concurrently

    References:
        TODO:
        http://stackoverflow.com/questions/6783194/background-thread-with-qthread-in-pyqt
    """

    def __init__(thread, thumb_path, img_path, img_size, thumbsize, qtindex,
                 view, offset, bbox_list, theta_list, interest_list):
        RUNNABLE_BASE.__init__(thread)
        thread.thumb_path = thumb_path
        thread.img_path = img_path
        thread.img_size = img_size
        thread.qtindex = qtindex
        thread.offset = offset
        thread.thumbsize = thumbsize
        thread.view = view
        thread.bbox_list = bbox_list
        thread.theta_list = theta_list
        thread.interest_list = interest_list

    def thumb_would_not_be_visible(thread):
        return view_would_not_be_visible(thread.view, thread.offset)

    def _run(thread):
        """ Compute thumbnail in a different thread """
        import vtool_ibeis as vt
        #time.sleep(.005)  # Wait a in case the user is just scrolling
        if thread.thumb_would_not_be_visible():
            return
        # Precompute info BEFORE reading the image (.0002s)
        dsize, new_verts_list = get_thread_thumb_info(
            thread.bbox_list, thread.theta_list, thread.thumbsize, thread.img_size)
        #time.sleep(.005)  # Wait a in case the user is just scrolling
        if thread.thumb_would_not_be_visible():
            return
        # -----------------
        # This part takes time, hopefully the user actually wants to see this
        # thumbnail.
        thumb = make_thread_thumb(thread.img_path, dsize, new_verts_list, thread.interest_list)
        if thread.thumb_would_not_be_visible():
            return
        vt.image.imwrite(thread.thumb_path, thumb)
        del thumb
        if thread.thumb_would_not_be_visible():
            return
        #print('[ThumbCreationThread] Thumb Written: %s' % thread.thumb_path)
        thread.qtindex.model().dataChanged.emit(thread.qtindex, thread.qtindex)
        #unregister_thread(thread.thumb_path)

    def run(thread):
        try:
            thread._run()
        except Exception as ex:
            ut.printex(ex, 'thread failed', tb=True)
            #raise

    #def __del__(self):
    #    print('About to delete creation thread')


# GRAVE:
#print('[APIItemDelegate] Request Thumb: rc=(%d, %d), nBboxes=%r' %
#      (qtindex.row(), qtindex.column(), len(bbox_list)))
#print('[APIItemDelegate] bbox_list = %r' % (bbox_list,))

def simple_thumbnail_widget():
    r"""
    Very simple example to test thumbnails

    CommandLine:
        python -m guitool_ibeis.api_thumb_delegate --test-simple_thumbnail_widget  --show --verbthumb
        python -m guitool_ibeis.api_thumb_delegate --test-simple_thumbnail_widget  --show --tb

    Example:
        >>> # GUI_DOCTEST
        >>> # xdoctest: +REQUIRES(--gui)
        >>> from guitool_ibeis.api_thumb_delegate import *  # NOQA
        >>> import guitool_ibeis
        >>> guitool_ibeis.ensure_qapp()  # must be ensured before any embeding
        >>> wgt = simple_thumbnail_widget()
        >>> ut.quit_if_noshow()
        >>> wgt.show()
        >>> guitool_ibeis.qtapp_loop(wgt, frequency=100)
    """
    import guitool_ibeis
    guitool_ibeis.ensure_qapp()
    col_name_list = ['rowid', 'image_name', 'thumb']
    col_types_dict = {
        'thumb': 'PIXMAP',
    }

    guitool_test_thumbdir = ut.ensure_app_resource_dir('guitool_ibeis', 'thumbs')
    ut.delete(guitool_test_thumbdir)
    ut.ensuredir(guitool_test_thumbdir)
    import vtool_ibeis as vt
    from os.path import join

    #imgname_list = sorted(ut.TESTIMG_URL_DICT.keys())
    imgname_list = ['carl.jpg', 'astro.png', 'patsy.jpg']
    imgname_list += ['doesnotexist.jpg']

    num_imgs = list(range(len(imgname_list)))
    # num_imgs = list(range(500))

    def thread_func(would_be, id_):
        from vtool_ibeis.fontdemo import get_text_test_img
        get_text_test_img(id_)

    def thumb_getter(id_, thumbsize=128):
        """ Thumb getters must conform to thumbtup structure """

        if id_ not in imgname_list:
            return {
                'fpath': id_ + '.jpg',
                'thread_func': thread_func,
                'main_func': lambda : (id_,),
            }
        #print(id_)
        if id_ == 'doesnotexist.jpg':
            return None
            img_path = None
            img_size = (100, 100)
        else:
            img_path = ut.grab_test_imgpath(id_, verbose=False)
            img_size = vt.open_image_size(img_path)
        thumb_path = join(guitool_test_thumbdir, ut.hashstr(str(img_path)) + '.jpg')
        if id_ == 'carl.jpg':
            bbox_list = [(10, 10, 200, 200)]
            theta_list = [0]
        elif id_ == 'astro.png':
            #bbox_list = [(10, 10, 200, 200)]
            bbox_list = [None]
            theta_list = [None]
        else:
            bbox_list = []
            theta_list = []
        interest_list = [False]
        thumbtup = (thumb_path, img_path, img_size, bbox_list, theta_list, interest_list)
        #print('thumbtup = %r' % (thumbtup,))
        return thumbtup
        #return None

    def imgname_getter(rowid):
        if rowid < len(imgname_list):
            return imgname_list[rowid]
        else:
            return str(rowid)

    col_getter_dict = {
        'rowid': num_imgs,
        'image_name': imgname_getter,
        'thumb': thumb_getter
    }
    col_ider_dict = {
        'thumb': 'image_name',
    }
    col_setter_dict = {}
    editable_colnames = []
    sortby = 'rowid'
    def get_thumb_size():
        return 128
    #get_thumb_size = lambda: 128  # NOQA
    col_width_dict = {}
    col_bgrole_dict = {}

    api = guitool_ibeis.CustomAPI(
        col_name_list, col_types_dict, col_getter_dict,
        col_bgrole_dict, col_ider_dict, col_setter_dict,
        editable_colnames, sortby, get_thumb_size, True, col_width_dict)
    headers = api.make_headers(tblnice='Utool Test Images')

    wgt = guitool_ibeis.APIItemWidget()
    wgt.change_headers(headers)
    wgt.resize(600, 400)
    #guitool_ibeis.qtapp_loop(qwin=wgt, ipy=ipy, frequency=loop_freq)
    return wgt


if __name__ == '__main__':
    """
    CommandLine:
        python -m guitool_ibeis.api_thumb_delegate
        python -m guitool_ibeis.api_thumb_delegate --allexamples
        python -m guitool_ibeis.api_thumb_delegate --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()
